import copy

import numpy as np
import plotly.graph_objects as go
import polars as pl
import streamlit as st
from matplotlib.path import Path
from pydantic import BaseModel, Field

from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.utils.ngio_io_caches import (
    get_ome_zarr_container,
    get_single_label_image,
)
from fractal_feature_explorer.utils.st_components import (
    selectbox_component,
    single_slider_component,
)
from streamlit.logger import get_logger

logger = get_logger(__name__)


def _show_point_info(
    point_dict: dict,
):
    st.write("Image URL: ", point_dict["image_url"])
    st.write("Label: ", point_dict["label"])
    st.write("Reference Label: ", point_dict["reference_label"])
    for key, value in point_dict.items():
        if key not in ["image_url", "label", "reference_label"]:
            st.write(f"{key}: ", value)


@st.dialog("Cell Preview")
def view_point(point: int, feature_df: pl.DataFrame) -> None:
    """
    View the point in the data frame
    """
    point_dict = feature_df.select("image_url", "label", "reference_label").to_dicts()[
        point
    ]
    logger.info(f"Opening point: {point_dict} in dialog")

    try:
        container = get_ome_zarr_container(
            point_dict["image_url"],
            mode="image",
        )
        image = container.get_image()
    except Exception as e:
        logger.error(f"Error opening image: {e}")
        st.error("Error opening image")
        _show_point_info(point_dict)
        return

    channels = container.image_meta.channel_labels
    if len(channels) > 1:
        channel = st.selectbox(
            label="Select channel",
            options=channels,
            index=0,
            help="Select the channel to display",
        )
        channel = channels.index(channel)
    else:
        channel = 0

    if container.is_3d:
        z_slice = st.slider(
            label="Select Z slice",
            min_value=0,
            max_value=image.dimensions.get("z"),
            value=0,
            help="Select the Z slice to display",
        )
    else:
        z_slice = 0

    if container.is_time_series:
        t_slice = st.slider(
            label="Select T slice",
            min_value=0,
            max_value=image.dimensions.get("t"),
            value=0,
            help="Select the T slice to display",
        )
    else:
        t_slice = 0

    show_label = st.toggle(
        label="Show Label", value=True, help="Show the label on the image"
    )

    with st.expander("Advanced Options", expanded=False):
        zoom_factor = st.slider(
            label="Zoom Factor",
            min_value=0.5,
            max_value=10.0,
            value=1.5,
            step=0.1,
            help="Zoom factor for the image",
        )

        level_path = st.selectbox(
            label="Select Level",
            options=container.levels_paths,
            index=0,
            help="Select the level to display",
        )

    try:
        image = get_single_label_image(
            image_url=point_dict["image_url"],
            ref_label=point_dict["reference_label"],
            label=int(point_dict["label"]),
            level_path=level_path,
            channel=channel,
            z_slice=z_slice,
            t_slice=t_slice,
            show_label=show_label,
            zoom_factor=zoom_factor,
        )
        st.image(image, use_container_width=True)
    except Exception as e:
        logger.error(f"Error opening image: {e}")
        st.error("Error opening image")
        _show_point_info(point_dict)
        return

    with st.expander("Infos", expanded=False):
        _show_point_info(point_dict)


class ScatterFilter(BaseModel):
    column_x: str
    column_y: str
    sel_x: list[float] = Field(default_factory=list)
    sel_y: list[float] = Field(default_factory=list)

    # model_config = ConfigDict(
    #   validate_assignment=True,
    # )

    def apply(self, feature_frame: FeatureFrame) -> FeatureFrame:
        """
        Filter the feature frame using the histogram filter
        """
        table = feature_frame.table.select(self.column_x, self.column_y).collect()
        mask = self.compute_selection_mask(table)
        filtered_table = feature_frame.table.filter(mask)
        return FeatureFrame(
            table=filtered_table,
            features=feature_frame.features,
            cathegorical=feature_frame.cathegorical,
            others=feature_frame.others,
        )

    def compute_selection_mask(self, feature_df: pl.DataFrame) -> np.ndarray:
        """
        Compute the selection mask for the feature frame using the histogram filter
        """
        assert len(self.sel_x) == len(self.sel_y), (
            "X and Y coordinates must be the same length"
        )
        if len(self.sel_x) == 0:
            return np.ones(len(feature_df), dtype=bool)

        x_column = feature_df[self.column_x].to_numpy()
        y_column = feature_df[self.column_y].to_numpy()
        poly_verts = np.column_stack((self.sel_x, self.sel_y))
        polygon = Path(poly_verts)
        pts = np.column_stack((x_column, y_column))
        mask = polygon.contains_points(pts)

        return mask

    def apply_to_df(self, feature_df: pl.DataFrame) -> pl.DataFrame:
        """
        Filter the feature frame using the histogram filter
        """
        mask = self.compute_selection_mask(feature_df)
        filtered_table = feature_df.filter(mask)
        return filtered_table


def scatter_filter_component(
    key: str,
    feature_frame: FeatureFrame,
) -> FeatureFrame:
    """
    Create a scatter filter for the feature frame
    And return the filtered feature frame
    """
    if len(feature_frame.features) < 2:
        error_msg = (
            "Not enough features found in the feature table. "
            "At least 2 features are required for the scatter filter."
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    col1, col2 = st.columns(2)
    features_columns = feature_frame.features
    with col1:
        x_column = selectbox_component(
            key=f"{key}:scatter_filter_x_column",
            label="Select **X-axis**",
            options=features_columns,
        )
        # remove x_column from the list of options
        _features_columns = copy.deepcopy(features_columns)
        _features_columns.remove(x_column)
        y_column = selectbox_component(
            key=f"{key}:scatter_filter_y_column",
            label="Select **Y-axis**",
            options=_features_columns,
        )
        feature_lf = feature_frame.table
        feature_df = feature_lf.select(
            x_column, y_column, "image_url", "label", "reference_label"
        ).collect()

    with col2:
        do_sampling = st.toggle(
            key=f"{key}:scatter_filter_sampling",
            label="Do sampling",
            value=True,
            help="If the number of points is too high, we will sample the points to display",
        )
        if do_sampling:
            if feature_df.height > 50000:
                default = 50000 / feature_df.height
            else:
                default = 1.0
            perc_samples = single_slider_component(
                key=f"{key}:scatter_filter_num_samples",
                label="Percentage of samples to display",
                min_value=0,
                max_value=1,
                default=default,
                help="Number of samples to display in the scatter plot.",
            )
        else:
            perc_samples = 1.0
            st.write("Number of points to display: ", feature_df.height)

        show_advanced_options = st.toggle(
            key=f"{key}:scatter_filter_advanced_options",
            label="Show advanced options",
            value=False,
            help="Show advanced options for the scatter plot",
        )
        if show_advanced_options:
            point_size = single_slider_component(
                key=f"{key}:scatter_filter_point_size",
                label="Point size",
                min_value=1,
                max_value=20,
                default=4,
                help="Size of the points in the scatter plot.",
            )
            point_size = int(point_size)
            opacity = single_slider_component(
                key=f"{key}:scatter_filter_opacity",
                label="Opacity",
                min_value=0.0,
                max_value=1.0,
                default=0.9,
                help="Opacity of the points in the scatter plot.",
            )
            opacity = float(opacity)
        else:
            point_size = 5
            opacity = 1.0

    if do_sampling:
        feature_df = feature_df.sample(n=int(feature_df.height * perc_samples), seed=0)

    fig = go.Figure()
    fig.update_layout(
        title="Scatter Plot",
        xaxis_title=x_column,
        yaxis_title=y_column,
        showlegend=True,
    )

    if f"{key}:state" in st.session_state:
        state = ScatterFilter.model_validate_json(st.session_state[f"{key}:state"])
        if len(state.sel_x) > 0:
            opacity_factor = 0.2
        else:
            opacity_factor = 1.0
    else:
        state = ScatterFilter(
            column_x=x_column,
            column_y=y_column,
        )
        opacity_factor = 1.0

    fig.add_trace(
        go.Scattergl(
            x=feature_df[x_column],
            y=feature_df[y_column],
            mode="markers",
            marker=dict(
                size=point_size,
                color="#1f77b4",
                opacity=opacity * opacity_factor,
            ),
            name="All Points",
        )
    )

    if len(state.sel_x) > 0:
        logger.info("Adding filtered points to the scatter plot")
        filtered_df = state.apply_to_df(feature_df)
        fig.add_trace(
            go.Scattergl(
                x=filtered_df[x_column],
                y=filtered_df[y_column],
                mode="markers",
                marker=dict(
                    size=point_size,
                    opacity=opacity,
                    color="#1f77b4",
                ),
                name="Selected Points",
            )
        )

        sel_x = state.sel_x + [state.sel_x[0]]
        sel_y = state.sel_y + [state.sel_y[0]]
        fig.add_trace(
            go.Scattergl(
                x=sel_x,
                y=sel_y,
                mode="lines+markers",  # Shows both lines and markers
                line=dict(color="#ff7f0e", width=2),
                marker=dict(size=8, opacity=1),
                name="Current Selection",
            )
        )
    fig.update_xaxes(showgrid=True)  # type: ignore
    fig.update_yaxes(showgrid=True)  # type: ignore

    event = st.plotly_chart(
        fig,
        key=f"{key}:scatter_plot",
        on_select="rerun",
        selection_mode=["points", "lasso"],
    )
    logger.info("Scatter plot created")
    selection = event.get("selection")
    if selection is not None:
        is_event_selection = (
            len(selection.get("box", [])) > 0 or len(selection.get("lasso", [])) > 0
        )
        is_click_selection = len(selection.get("point_indices", [])) > 0
        if is_event_selection:
            if len(selection.get("lasso", [])) > 0:
                if st.button(
                    "Confirm selection", key=f"{key}:confirm_selection", icon="✅"
                ):
                    scatter_state = ScatterFilter(
                        column_x=x_column,
                        column_y=y_column,
                        sel_x=selection.get("lasso", [])[0].get("x", []),
                        sel_y=selection.get("lasso", [])[0].get("y", []),
                    )
                    st.session_state[f"{key}:state"] = scatter_state.model_dump_json()
                    logger.info(f"Adding scatter filter state: {scatter_state}")
                    st.rerun()
            else:
                if f"{key}:state" in st.session_state:
                    logger.info("Removing scatter filter state")
                    del st.session_state[f"{key}:state"]
                    st.rerun()

        elif is_click_selection:
            logger.info("Click selection on the scatter plot")
            view_point(
                point=selection.get("point_indices", [])[0],
                feature_df=feature_df,
            )

    st.session_state[f"{key}:type"] = "scatter"
    if f"{key}:state" in st.session_state:
        scatter_state = ScatterFilter.model_validate_json(
            st.session_state[f"{key}:state"]
        )
        feature_frame = scatter_state.apply(feature_frame=feature_frame)
        logger.info(
            f"Scatter filter applied: {scatter_state.column_x} [{scatter_state.sel_x}, {scatter_state.sel_y}]"
        )
        return feature_frame
    scatter_state = ScatterFilter(
        column_x=x_column,
        column_y=y_column,
    )
    st.session_state[f"{key}:state"] = scatter_state.model_dump_json()
    return feature_frame



from pathlib import Path
from pydantic import BaseModel, ConfigDict, AfterValidator
import toml
import streamlit as st
from typing import Literal
from typing import Annotated

from streamlit.logger import get_logger
import os

logger = get_logger(__name__)

CONFIG_PATH = os.getenv(
    "FRACTAL_FEATURE_EXPLORER_CONFIG",
    (Path.home() / ".fractal_feature_explorer" / "config.toml"),
)

DEFAULT_CONFIG_PATH = Path(__file__).parent / "resources" / "config.toml"


def remove_trailing_slash(value: str) -> str:
    return value.rstrip("/")


class LocalConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    deployment_type: Literal["local"]
    fractal_data_urls: list[Annotated[str, AfterValidator(remove_trailing_slash)]]
    allow_local_paths: bool = True


class ProductionConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    deployment_type: Literal["production"]
    allow_local_paths: Literal[False] = False
    fractal_data_url: Annotated[str, AfterValidator(remove_trailing_slash)]
    fractal_backend_url: Annotated[str, AfterValidator(remove_trailing_slash)]
    fractal_frontend_url: Annotated[str, AfterValidator(remove_trailing_slash)]
    fractal_cookie_name: str = "fastapiusersauth"


def _init_local_config() -> LocalConfig:
    """
    Initialize the local configuration with default values.
    """
    config_path = Path(CONFIG_PATH)
    allow_saving_config = input(
        f"Do you want to create a default configuration file at {config_path.as_posix()}? (y/n): "
    )
    for _ in range(3):
        if allow_saving_config.lower() == "y" or allow_saving_config.lower() == "n":
            break
        else:
            allow_saving_config = input(
                f"{allow_saving_config} is not a valid input. Please enter 'y' or 'n': "
            )
    else:
        raise ValueError(
            "Too many invalid inputs. Exiting without saving the configuration."
        )

    with open(DEFAULT_CONFIG_PATH, "r") as f:
        default_config = toml.load(f)

    local_config = LocalConfig(**default_config)
    if allow_saving_config.lower() == "y":
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, "w") as f:
            toml.dump(local_config.model_dump(), f)
        logger.info(f"Default configuration saved to {config_path.as_posix()}.")

    return local_config


@st.cache_data
def get_config() -> LocalConfig | ProductionConfig:
    """
    Get the configuration for the Fractal Explorer.
    """
    config_path = Path(CONFIG_PATH)

    if config_path.exists():
        config_data = toml.load(config_path)
        key = "deployment_type"
        if key not in config_data.keys():
            raise ValueError(f"Missing {key=} in {config_path=}.")
        elif config_data[key] == "local":
            config = LocalConfig(**config_data)
            logger.info(f"Local configuration read from {config_path.as_posix()}.")
        elif config_data[key] == "production":
            config = ProductionConfig(**config_data)
            logger.info(f"Production configuration read from {config_path.as_posix()}.")
        else:
            raise ValueError(
                f"Invalid {key=} in {config_path=}. Expected 'local' or 'production'."
            )
    else:
        logger.warning(
            f"Config file {config_path} does not exist; "
            "using default local configuration."
        )
        config = _init_local_config()
    logger.debug(f"{config=}")
    return config



import io

import polars as pl
import streamlit as st

from fractal_feature_explorer.pages.filters_page import (
    apply_filters,
    build_feature_frame,
)
from fractal_feature_explorer.utils import Scope
from fractal_feature_explorer.authentication import verify_authentication


def table_to_csv_buffer(table: pl.LazyFrame) -> io.BytesIO:
    """
    Convert a Polars DataFrame to a CSV buffer.
    """
    _table = table.collect()
    buffer = io.BytesIO()
    _table.write_csv(buffer)
    buffer.seek(0)
    return buffer


def table_to_parquet_buffer(table: pl.LazyFrame) -> io.BytesIO:
    """
    Convert a Polars DataFrame to a Parquet buffer.
    """
    _table = table.collect()
    buffer = io.BytesIO()
    _table.write_parquet(buffer)
    buffer.seek(0)
    return buffer


def main():
    verify_authentication()
    feature_table = st.session_state.get(f"{Scope.DATA}:feature_table", None)
    feature_table_name = st.session_state.get(f"{Scope.DATA}:feature_table_name", "")
    if feature_table is None:
        st.warning(
            "No feature table found in session state. Please make sure to run the setup page first."
        )
        st.stop()

    export_format = st.pills(
        label="Select Export Format",
        options=["CSV", "Parquet"],
        default=None,
        help="Select the format to export the Table.",
    )

    if export_format is None:
        st.warning("Please select an export format to download the filtered data.")
        st.stop()

    if st.toggle(
        label="Apply Filters",
        value=True,
        help="Toggle to apply filters to the feature table.",
    ):
        feature_frame = build_feature_frame(
            feature_table,
        )
        feature_table = apply_filters(feature_frame=feature_frame).table

    if export_format == "CSV":
        file = table_to_csv_buffer(feature_table)
        st.download_button(
            label="Download Filtered Data as CSV",
            data=file,
            file_name=f"{feature_table_name}_filtered.csv",
            mime="text/csv",
            on_click="ignore",
            icon="📥",
        )
    elif export_format == "Parquet":
        file = table_to_parquet_buffer(feature_table)
        st.download_button(
            label="Download Filtered Data as Parquet",
            data=file,
            file_name=f"{feature_table_name}_filtered.parquet",
            mime="application/octet-stream",
            on_click="ignore",
            icon="📥",
        )


if __name__ == "__main__":
    main()



"""Fractal Explorer Dashboard"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("fractal-feature-explorer")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"
__author__ = "Lorenzo Cerrone"
__email__ = "lorenzo.cerrone@uzh.ch"



import streamlit as st
from pydantic import BaseModel, ConfigDict

from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.utils.st_components import (
    multiselect_component,
)


class ColumnsFilter(BaseModel):
    features: list[str]
    cathegorical: list[str]
    others: list[str]

    model_config = ConfigDict(
        validate_assignment=True,
    )

    def apply(self, feature_frame: FeatureFrame) -> FeatureFrame:
        """
        Filter the feature frame using the columns filter
        """
        all_columns = set(
            self.features + self.cathegorical + self.others + feature_frame.protected
        )
        filtered_table = feature_frame.table.select(
            all_columns,
        )
        return FeatureFrame(
            table=filtered_table,
            features=self.features,
            cathegorical=self.cathegorical,
            others=self.others,
        )


def columns_filter_component(
    key: str,
    feature_frame: FeatureFrame,
) -> FeatureFrame:
    """
    Filter the feature table to only include the specified columns.
    """
    features = feature_frame.features
    selected_features = multiselect_component(
        key=f"{key}:features_filter",
        label="Select features to include in the feature table",
        options=features,
    )

    cathegorical = set(feature_frame.cathegorical) - set(feature_frame.protected)
    selected_categorical = multiselect_component(
        key=f"{key}:categorical_filter",
        label="Select categorical features to include in the feature table",
        options=list(cathegorical),
    )

    others = set(feature_frame.others) - set(feature_frame.protected)
    selected_others = multiselect_component(
        key=f"{key}:others_filter",
        label="Select other features to include in the feature table",
        options=list(others),
    )
    selected_others = list(set(selected_others) - set(selected_categorical))

    all_columns = set(
        selected_features
        + selected_categorical
        + selected_others
        + feature_frame.protected
    )
    filtered_table = feature_frame.table.select(all_columns)

    feature_frame = FeatureFrame(
        table=filtered_table,
        features=selected_features,
        cathegorical=selected_categorical,
        others=selected_others,
    )

    st.session_state[f"{key}:type"] = "columns"
    st.session_state[f"{key}:state"] = ColumnsFilter(
        features=selected_features,
        cathegorical=selected_categorical,
        others=selected_others,
    ).model_dump_json()
    return feature_frame



from fractal_feature_explorer.utils.ngio_io_caches import _url_belongs_to_base


def test_url_belongs_to_base():
    base_url = "https://example.com/base/"
    test_urls = [
        "https://example.com/base/resource",
        "https://example.com/base/resource/subresource",
        "https://example.com/base/resource/subresource?query=param#fragment",
    ]

    for url in test_urls:
        assert _url_belongs_to_base(url, base_url)

    negative_test_urls = [
        "https://example.com/other/resource",
        "http://example.com/base/resource",
    ]

    for url in negative_test_urls:
        assert not _url_belongs_to_base(url, base_url)



import streamlit as st

from fractal_feature_explorer.utils import get_and_validate_store, is_http_url
from streamlit.logger import get_logger

logger = get_logger(__name__)


def sanify_http_url(url: str) -> str:
    """Sanitize the URL by removing the trailing slash."""
    url = url.replace(" ", "%20")

    if url.endswith("/"):
        url = url[:-1]
    return url


def sanify_path_url(url: str) -> str:
    """Sanitize a local path URL."""
    # Remove any leading spaces
    url = url.lstrip(" ")

    # Remove string quotes
    url = url.lstrip('"').rstrip('"')
    url = url.lstrip("'").rstrip("'")
    return str(url)


def sanify_and_validate_url(url: str) -> str | None:
    """Sanitize the URL by removing the trailing slash."""
    if len(url) == 0:
        st.error("URL is empty.")
        logger.error("URL is empty.")
        return None

    if is_http_url(url):
        url = sanify_http_url(url)
    else:
        url = sanify_path_url(url)

    store = get_and_validate_store(url)
    if store is None:
        return None

    return url


def plate_name_from_url(plate_url: str) -> str:
    """Get the plate name from the URL."""
    return plate_url.rsplit("/", 1)[-1]


def extras_from_url(image_url: str) -> dict[str, str]:
    """Get the extras from the URL."""
    *_, plate_name, row, column, path_in_well = image_url.split("/")
    return {
        "plate_name": plate_name,
        "row": row,
        "column": column,
        "path_in_well": path_in_well,
    }



from pathlib import Path
import shutil

config_path = Path("tests/configs/local.toml")
if not config_path.exists():
    raise FileNotFoundError("Test configuration file not found.")

# Copy the test to the expected location in the home directory
home = Path.home()
test_config_path = home / ".fractal_feature_explorer" / "config.toml"
test_config_path.parent.mkdir(parents=True, exist_ok=True)
shutil.copy(config_path, test_config_path)



import streamlit as st
import fractal_feature_explorer
import ngio
from fractal_feature_explorer.config import get_config
from pathlib import Path


def main():
    icon_path = Path(__file__).parent / "resources" / "fractal_favicon.png"
    logo_path = Path(__file__).parent / "resources" / "fractal_logo.png"

    st.set_page_config(
        layout="wide",
        page_title="Fractal Plate Explorer",
        page_icon=icon_path,
    )
    t_col1, t_col2 = st.columns([1, 5])
    with t_col1:
        st.image(logo_path, width=100)
    with t_col2:
        st.title("Fractal Explorer")

    footer_style = """
    <style>
        .footer {
            position: fixed;
            bottom: 0;
            left: 0;
            width: 100%;
            background-color: #f0f2f6;
            text-align: center;
            padding: 10px;
            font-size: 0.9em;
            color: #6c757d;
        }
    </style>
    """

    footer_content = f"""
    <div class='footer'>
        Fractal Explorer Version: {fractal_feature_explorer.__version__} |
        ngio Version: {ngio.__version__} |
        © Copyright 2025 University of Zurich (see LICENSE file for details)
    </div>
    """
    st.markdown(f"{footer_style}{footer_content}", unsafe_allow_html=True)

    setup_page = st.Page(
        "pages/setup_page/setup_page.py",
        title="OME-Zarr Setup",
        icon=":material/settings:",
    )
    filter_page = st.Page(
        "pages/filters_page/filters_page.py",
        title="Features Filters",
        icon=":material/filter:",
    )
    explore_page = st.Page(
        "pages/explore_page/explore_page.py", title="Explore", icon=":material/search:"
    )
    export_page = st.Page(
        "pages/export_page.py", title="Export", icon=":material/download:"
    )

    pages = [setup_page, filter_page, explore_page, export_page]
    config = get_config()
    if config.deployment_type == "production":
        user_info_page = st.Page("pages/info.py", title="Info", icon=":material/info:")
        pages.append(user_info_page)

    pg = st.navigation(pages)
    pg.run()


if __name__ == "__main__":
    main()



import polars as pl
import streamlit as st
from streamlit.logger import get_logger

from fractal_feature_explorer.pages.explore_page._heat_map_plot import (
    heat_map_component,
)
from fractal_feature_explorer.pages.explore_page._scatter_plot import (
    scatter_plot_component,
)
from fractal_feature_explorer.pages.filters_page import (
    apply_filters,
    build_feature_frame,
)
from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.utils import Scope, invalidate_session_state
from fractal_feature_explorer.authentication import verify_authentication

logger = get_logger(__name__)


def _find_unique_name(keys: list[str], prefix: str) -> str:
    """
    Find a unique name for the filter
    """
    i = 1
    while True:
        name = f"{prefix} {i}"
        if name not in keys:
            return name
        i += 1


def add_plot() -> None:
    """
    Dynamically add a plot to the explorer page
    """
    if f"{Scope.EXPLORE}:plots_dict" not in st.session_state:
        st.session_state[f"{Scope.EXPLORE}:plots_dict"] = {}

    plot_type = st.pills(
        label="Plot Type",
        options=["Scatter Plot", "Heat Map"],
        default="Scatter Plot",
        key=f"{Scope.EXPLORE}:plot_type",
        selection_mode="single",
        help="Select the type of plot to add.",
    )

    if st.button("Add New Plot", key=f"{Scope.EXPLORE}:add_plot_button"):
        if plot_type == "Scatter Plot":
            name = _find_unique_name(
                st.session_state[f"{Scope.EXPLORE}:plots_dict"].keys(),
                "Scatter Plot",
            )
            key = f"{Scope.EXPLORE}:{name}_scatter_plot"
            st.session_state[f"{Scope.EXPLORE}:plots_dict"][name] = (
                key,
                scatter_plot_component,
            )
            logger.info(f"New Scatter Plot added: {name}")
            st.rerun()
        elif plot_type == "Heat Map":
            name = _find_unique_name(
                st.session_state[f"{Scope.EXPLORE}:plots_dict"].keys(),
                "Heat Map",
            )
            key = f"{Scope.EXPLORE}:{name}_heat_map"
            st.session_state[f"{Scope.EXPLORE}:plots_dict"][name] = (
                key,
                heat_map_component,
            )
            logger.info(f"New Heat Map added: {name}")
            st.rerun()

    return None


def display_plots(feature_frame: FeatureFrame) -> FeatureFrame:
    """
    Display the plots in the feature table
    """
    plot_list = st.session_state[f"{Scope.EXPLORE}:plots_dict"]

    for name, (plot_key, plot_component) in plot_list.items():
        st.markdown(
            f"""
            ### {name}
            """
        )
        try:
            plot_component(
                key=plot_key,
                feature_frame=feature_frame,
            )
        except Exception as e:
            error_msg = f"Error displaying plot {name}: {e}"
            logger.error(error_msg)
            st.error(error_msg)

        col1, col2 = st.columns(2)
        with col1:
            if st.button("Reset Plot", key=f"{plot_key}:reset_plot_button", icon="🔄"):
                invalidate_session_state(plot_key)
                st.rerun()
        with col2:
            if st.button(
                "Delete Plot", key=f"{plot_key}:delete_plot_button", icon="🚮"
            ):
                invalidate_session_state(plot_key)
                del plot_list[name]
                st.session_state[f"{Scope.EXPLORE}:plots_dict"] = plot_list
                st.rerun()

    return feature_frame


def feature_explore_manager(
    feature_table: pl.LazyFrame, table_name: str, skip_filters: bool = True
) -> FeatureFrame:
    """
    Setup the feature table for the dashboard.
    """
    st.markdown(
        f"""
        ## Explore the feature table
        
        Table Name: **{table_name}**
        """
    )
    feature_frame = build_feature_frame(feature_table)
    if not skip_filters:
        feature_frame = apply_filters(feature_frame)

    col1, _ = st.columns(2)
    with col1:
        add_plot()

    feature_frame = display_plots(feature_frame)
    return feature_frame


def main():
    verify_authentication()
    with st.sidebar:
        with st.expander("Advanced Options", expanded=False):
            skip_filters = st.toggle(
                label="Skip Filters",
                key=f"{Scope.EXPLORE}:apply_filters",
                value=False,
            )

            if st.button(
                "Reset Explore Page",
                key=f"{Scope.EXPLORE}:reset_explore_page",
                icon="🔄",
                help="Reset the explore page state. This will clear all plots.",
            ):
                invalidate_session_state(f"{Scope.EXPLORE}")
                st.rerun()

    feature_table = st.session_state.get(f"{Scope.DATA}:feature_table", None)
    feature_table_name = st.session_state.get(f"{Scope.DATA}:feature_table_name", "")

    if feature_table is None:
        st.warning(
            "No feature table found in session state. Please make sure to run the setup page first."
        )
        st.stop()

    feature_explore_manager(
        feature_table=feature_table,
        table_name=feature_table_name,
        skip_filters=skip_filters,
    )
    logger.info("Explore page loading complete")


if __name__ == "__main__":
    main()



from typing import Literal

import streamlit as st
from pydantic import BaseModel, ConfigDict, model_validator
from streamlit.logger import get_logger

logger = get_logger(__name__)


class PillsState(BaseModel):
    options: list
    default: list

    model_config = ConfigDict(
        validate_assignment=True,
    )

    @model_validator(mode="after")
    def check_model(self):
        # default should be a subset of options
        # if not set default to the intersection of options and default
        if not set(self.default).issubset(set(self.options)):
            self.default = list(set(self.options).intersection(set(self.default)))
        return self


def pills_component(
    key: str,
    label: str,
    options: list,
    selection_mode: Literal["multi", "single"] = "multi",
    help: str | None = None,
) -> list:
    """Wrapper for the streamlit.pills component.

    This function handles the session state for the pills component.
    A unique key is required for each pills component.
    """
    if f"{key}_state_model" not in st.session_state:
        pills_model = PillsState(
            options=options,
            default=options,
        )
        st.session_state[f"{key}_state_model"] = pills_model.model_dump_json()
    else:
        pills_model = PillsState.model_validate_json(
            st.session_state[f"{key}_state_model"]
        )
        pills_model.options = options
        if len(pills_model.default) == 0 and len(options) > 0:
            pills_model.default = [options[0]]

    def _update_session_state():
        pills_model.default = st.session_state[key]
        st.session_state[f"{key}_state_model"] = pills_model.model_dump_json()

    selected_options = st.pills(
        label=label,
        options=pills_model.options,
        default=pills_model.default,
        selection_mode=selection_mode,
        help=help,
        key=key,
        on_change=_update_session_state,
    )

    if selected_options is None:
        selected_options = []
    if not isinstance(selected_options, list):
        selected_options = [selected_options]

    return selected_options


class SelectBoxState(BaseModel):
    options: list
    index: int = 0

    model_config = ConfigDict(
        validate_assignment=True,
    )

    @model_validator(mode="after")
    def check_model(self):
        # index should be less than the length of options
        if len(self.options) == 0:
            return self

        if self.index >= len(self.options):
            self.index = 0
        return self


def selectbox_component(
    key: str,
    label: str,
    options: list[str],
    help: str | None = None,
) -> str:
    """Wrapper for the streamlit.selectbox component.
    This function handles the session state for the selectbox component.
    A unique key is required for each selectbox component.
    """

    if len(options) == 0:
        raise ValueError(
            "Selectbox options are empty. Selectbox cannot be created with an empty list."
        )

    if f"{key}_state_model" not in st.session_state:
        selectbox_model = SelectBoxState(
            options=options,
        )
        st.session_state[f"{key}_state_model"] = selectbox_model.model_dump_json()
    else:
        selectbox_model = SelectBoxState.model_validate_json(
            st.session_state[f"{key}_state_model"]
        )
        selectbox_model.options = options

    def _update_session_state():
        selected_option = st.session_state[key]
        selectbox_model.index = selectbox_model.options.index(selected_option)
        st.session_state[f"{key}_state_model"] = selectbox_model.model_dump_json()

    selected_option = st.selectbox(
        label=label,
        options=selectbox_model.options,
        index=selectbox_model.index,
        help=help,
        key=key,
        on_change=_update_session_state,
    )

    return selected_option


class MultiSelectState(BaseModel):
    options: list
    default: list

    model_config = ConfigDict(
        validate_assignment=True,
    )

    @model_validator(mode="after")
    def check_model(self):
        # default should be a subset of options
        # if not set default to the intersection of options and default
        if not set(self.default).issubset(set(self.options)):
            self.default = self.options
        return self


def multiselect_component(
    key: str,
    label: str,
    options: list,
    help: str | None = None,
) -> list:
    """Wrapper for the streamlit.multiselect component.
    This function handles the session state for the multiselect component.
    A unique key is required for each multiselect component.
    """
    if f"{key}_state_model" not in st.session_state:
        multiselect_model = MultiSelectState(
            options=options,
            default=options,
        )
        st.session_state[f"{key}_state_model"] = multiselect_model.model_dump_json()
    else:
        multiselect_model = MultiSelectState.model_validate_json(
            st.session_state[f"{key}_state_model"]
        )
        multiselect_model.options = options
        if len(multiselect_model.default) == 0 and len(options) > 0:
            multiselect_model.default = [options[0]]

    def _update_session_state():
        multiselect_model.default = st.session_state[key]
        st.session_state[f"{key}_state_model"] = multiselect_model.model_dump_json()

    selected_options = st.multiselect(
        label=label,
        options=multiselect_model.options,
        default=multiselect_model.default,
        help=help,
        key=key,
        on_change=_update_session_state,
    )

    return selected_options


class SliderState(BaseModel):
    min_value: float
    max_value: float
    value: float | tuple[float, float]

    model_config = ConfigDict(
        validate_assignment=True,
    )

    @model_validator(mode="after")
    def check_model(self):
        # value should be between min_value and max_value
        if isinstance(self.value, tuple):
            if self.value[0] < self.min_value or self.value[1] > self.max_value:
                self.value = (self.min_value, self.max_value)
        else:
            if self.value < self.min_value or self.value > self.max_value:
                self.value = (self.min_value + self.max_value) / 2
        return self


def _slider_component(
    key: str,
    label: str,
    min_value: float,
    max_value: float,
    value: float | tuple[float, float],
    help: str | None = None,
) -> float | tuple[float, float]:
    """Wrapper for the streamlit.slider component.
    This function handles the session state for the slider component.
    A unique key is required for each slider component.
    """
    if f"{key}_state_model" not in st.session_state:
        slider_model = SliderState(
            min_value=min_value,
            max_value=max_value,
            value=value,
        )
        st.session_state[f"{key}_state_model"] = slider_model.model_dump_json()
    else:
        slider_model = SliderState.model_validate_json(
            st.session_state[f"{key}_state_model"]
        )
        slider_model.min_value = min_value
        slider_model.max_value = max_value

    def _update_session_state():
        slider_model.value = st.session_state[key]
        st.session_state[f"{key}_state_model"] = slider_model.model_dump_json()

    selected_option = st.slider(
        label=label,
        min_value=slider_model.min_value,
        max_value=slider_model.max_value,
        value=slider_model.value,
        help=help,
        key=key,
        on_change=_update_session_state,
    )

    return selected_option


def double_slider_component(
    key: str,
    label: str,
    min_value: float,
    max_value: float,
    help: str | None = None,
) -> tuple[float, float]:
    """Wrapper for the streamlit.slider component.
    This function handles the session state for the slider component.
    A unique key is required for each slider component.
    """
    range_values = _slider_component(
        key=key,
        label=label,
        min_value=min_value,
        max_value=max_value,
        value=(min_value, max_value),
        help=help,
    )
    if not isinstance(range_values, tuple):
        raise ValueError(
            f"Expected a tuple for the range values, got {type(range_values)}"
        )

    return range_values


def single_slider_component(
    key: str,
    label: str,
    min_value: float,
    max_value: float,
    default: float | None = None,
    help: str | None = None,
) -> float:
    """Wrapper for the streamlit.slider component.
    This function handles the session state for the slider component.
    A unique key is required for each slider component.
    """
    default_value = default if default is not None else (min_value + max_value) / 2
    value = _slider_component(
        key=key,
        label=label,
        min_value=min_value,
        max_value=max_value,
        value=default_value,
        help=help,
    )
    if isinstance(value, tuple):
        raise ValueError(f"Expected a single value for the slider, got {type(default)}")
    return value


class NumberInputState(BaseModel):
    min_value: float | int
    max_value: float | int
    value: float | int

    model_config = ConfigDict(
        validate_assignment=True,
    )

    @model_validator(mode="after")
    def check_model(self):
        # value should be between min_value and max_value
        if self.value < self.min_value or self.value > self.max_value:
            _value = (self.min_value + self.max_value) / 2
            if isinstance(self.value, int):
                _value = int(_value)
            self.value = _value
        return self


def number_input_component(
    key: str,
    label: str,
    min_value: float | int,
    max_value: float | int,
    value: float | int,
    step: float | int = 1,
    help: str | None = None,
) -> float | int:
    """Wrapper for the streamlit.number_input component.
    This function handles the session state for the number input component.
    A unique key is required for each number input component.
    """
    if f"{key}_state_model" not in st.session_state:
        number_input_model = NumberInputState(
            min_value=min_value,
            max_value=max_value,
            value=value,
        )
        st.session_state[f"{key}_state_model"] = number_input_model.model_dump_json()
    else:
        number_input_model = NumberInputState.model_validate_json(
            st.session_state[f"{key}_state_model"]
        )
        number_input_model.min_value = min_value
        number_input_model.max_value = max_value

    def _update_session_state():
        number_input_model.value = st.session_state[key]
        st.session_state[f"{key}_state_model"] = number_input_model.model_dump_json()

    selected_option = st.number_input(
        label=label,
        min_value=number_input_model.min_value,
        max_value=number_input_model.max_value,
        value=number_input_model.value,
        help=help,
        key=key,
        step=step,
        on_change=_update_session_state,
    )

    return selected_option



import copy

import plotly.express as px
import streamlit as st
from streamlit.logger import get_logger
from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.pages.filters_page._scatter_filter import view_point
from fractal_feature_explorer.utils.st_components import (
    selectbox_component,
    single_slider_component,
)

logger = get_logger(__name__)


def scatter_plot_component(
    key: str,
    feature_frame: FeatureFrame,
) -> None:
    """
    Create a scatter plot for the feature frame
    And return the filtered feature frame
    """
    if len(feature_frame.features) < 2:
        error_msg = (
            "Not enough features found in the feature table. "
            "At least 2 features are required for the scatter filter."
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    features_columns = feature_frame.features
    x_column = selectbox_component(
        key=f"{key}:scatter_plot_x_column",
        label="**X-axis**",
        options=features_columns,
    )
    # remove x_column from the list of options
    _features_columns = copy.deepcopy(features_columns)
    _features_columns.remove(x_column)
    y_column = selectbox_component(
        key=f"{key}:scatter_plot_y_column",
        label="**Y-axis**",
        options=_features_columns,
    )
    columns_needed = {x_column, y_column, "image_url", "label", "reference_label"}

    with st.expander("Advanced Options", expanded=False):
        do_sampling = st.toggle(
            key=f"{key}:scatter_plot_advanced_options",
            label="Randomly sample points",
            value=True,
        )

        if do_sampling:
            perc_samples = single_slider_component(
                key=f"{key}:scatter_plot_num_samples",
                label="Percentage of samples to display",
                min_value=0,
                max_value=1,
                default=0.1,
                help="Number of samples to display in the scatter plot.",
            )
        else:
            perc_samples = 1.0

        # color
        possible_color_columns = (
            ["--No Color--"]
            + feature_frame.cathegorical
            + feature_frame.protected
            + feature_frame.features
        )
        possible_color_columns = copy.deepcopy(possible_color_columns)
        for col in columns_needed:
            if col in possible_color_columns:
                possible_color_columns.remove(col)

        color_column = selectbox_component(
            key=f"{key}:scatter_plot_color_column",
            label="**Color**",
            options=possible_color_columns,
            help="Select the column to color the points by.",
        )
        if color_column != "--No Color--":
            columns_needed.add(color_column)
        else:
            color_column = None

        # size
        possible_size_columns = ["--No Size--"] + feature_frame.features
        possible_size_columns = copy.deepcopy(possible_size_columns)
        for col in columns_needed:
            if col in possible_size_columns:
                possible_size_columns.remove(col)
        size_column = selectbox_component(
            key=f"{key}:scatter_plot_size_column",
            label="**Size**",
            options=possible_size_columns,
        )
        if size_column != "--No Size--":
            columns_needed.add(size_column)
        else:
            size_column = None

        # marginal x
        marginal_x = selectbox_component(
            key=f"{key}:scatter_plot_marginal_x",
            label="**Marginal X**",
            options=["--No Marginal--", "histogram", "violin"],
        )
        if marginal_x == "--No Marginal--":
            marginal_x = None

        # marginal y
        marginal_y = selectbox_component(
            key=f"{key}:scatter_plot_marginal_y",
            label="**Marginal Y**",
            options=["--No Marginal--", "histogram", "violin"],
        )
        if marginal_y == "--No Marginal--":
            marginal_y = None

    feature_lf = feature_frame.table
    feature_df = feature_lf.select(columns_needed).collect()
    if do_sampling:
        feature_df = feature_df.sample(
            n=int(perc_samples * feature_df.shape[0]), seed=0
        )

    fig = px.scatter(
        data_frame=feature_df,
        x=x_column,
        y=y_column,
        color=color_column,
        size=size_column,
        marginal_x=marginal_x,
        marginal_y=marginal_y,
    )
    fig.update_xaxes(showgrid=True)  # type: ignore
    fig.update_yaxes(showgrid=True)  # type: ignore

    event = st.plotly_chart(fig, key=f"{key}:scatter_plot", on_select="rerun")
    logger.info("Scatter plot created")
    selection = event.get("selection")
    if selection is not None:
        is_event_selection = (
            len(selection.get("box", [])) > 0 or len(selection.get("lasso", [])) > 0
        )
        is_click_selection = (
            not is_event_selection and len(selection.get("point_indices", [])) > 0
        )
        if is_click_selection:
            view_point(
                point=selection.get("point_indices", [])[0],
                feature_df=feature_df,
            )



import urllib3.util
from fractal_feature_explorer.config import get_config, ProductionConfig
import urllib3
import streamlit as st
from fractal_feature_explorer.utils import Scope
from streamlit.logger import get_logger


logger = get_logger(__name__)


def _verify_authentication(config: ProductionConfig):
    logger.debug("Enter _verify_authentication.")

    current_email = st.session_state.get(f"{Scope.PRIVATE}:fractal-email", None)
    current_token = st.session_state.get(f"{Scope.PRIVATE}:fractal-token", None)

    if None not in (current_email, current_token):
        logger.info("User session is already authenticated.")
    else:
        logger.info("Proceed with user authentication.")
        # Extract cookie and token from browser
        try:
            cookie = next(
                _cookie.strip()
                for _cookie in st.context.headers["cookie"].split(";")
                if _cookie.strip().startswith(config.fractal_cookie_name)
            )
            token = cookie.split("=")[1]
        except StopIteration:
            msg = "Could not find the expected cookie."
            logger.info(msg)
            raise ValueError(msg)
        except IndexError:
            msg = "Invalid cookie."
            raise ValueError(msg)
        # Get user information from Fractal backend
        logger.info("Now obtain user information.")
        current_user_url = f"{config.fractal_backend_url}/auth/current-user/"

        response = urllib3.request(
            "GET",
            current_user_url,
            headers={"Authorization": f"Bearer {token}"},
        )
        if response.status == 200:
            email_address = response.json()["email"]
            st.session_state[f"{Scope.PRIVATE}:fractal-email"] = email_address
            st.session_state[f"{Scope.PRIVATE}:fractal-token"] = token
            logger.info("Obtained user information.")
        else:
            msg = f"Could not obtain Fractal user information from {current_user_url}."
            logger.info(msg)
            raise ValueError(msg)


def verify_authentication():
    config = get_config()
    if config.deployment_type == "local":
        return

    try:
        _verify_authentication(config)
    except Exception as e:
        logger.info(f"Authentication failed. Original error: {str(e)}.")
        login_url = f"{config.fractal_frontend_url}/auth/login"
        MSG = (
            "You are not authenticated as a Fractal user. "
            f"Please login at {login_url} and "
            "then refresh the current page."
        )
        st.error(MSG)
        st.stop()



from fractal_feature_explorer.pages.filters_page.filters_page import (
    apply_filters,
    build_feature_frame,
    feature_filters_manger,
)

__all__ = ["feature_filters_manger", "apply_filters", "build_feature_frame"]



import asyncio

import polars as pl
import streamlit as st

from fractal_feature_explorer.utils.common import Scope
from fractal_feature_explorer.utils.ngio_io_caches import (
    get_ome_zarr_plate,
)
from fractal_feature_explorer.utils.st_components import (
    pills_component,
    selectbox_component,
)


from streamlit.logger import get_logger
from fractal_feature_explorer.pages.setup_page._plate_advanced_selection import (
    advanced_plate_selection_component,
)
from fractal_feature_explorer.pages.setup_page._tables_io import (
    list_images_tables,
    list_plate_tables,
    collect_feature_table_from_images,
    collect_feature_table_from_plates,
)
from fractal_feature_explorer.pages.setup_page._utils import (
    sanify_and_validate_url,
    extras_from_url,
)

logger = get_logger(__name__)

# ====================================================================
#
# Zarr URL Setup:
# allow the user to add Zarr URLs to the DataFrame
#
# ====================================================================


def user_plate_url_input_component():
    """Create a widget for inputting plate URLs."""
    st.markdown("## Input Plate URLs")
    global_urls = st.session_state.get(f"{Scope.SETUP}:zarr_urls", [])
    logger.info(f"Global URLs: {global_urls}")

    if f"{Scope.SETUP}:plate_setup:urls" not in st.session_state:
        st.session_state[f"{Scope.SETUP}:plate_setup:urls"] = set()

    new_url = st.text_input("Plate URL")
    if st.button("Add Plate URL", icon="➕"):
        # Validate the URL
        new_url = sanify_and_validate_url(new_url)
        if new_url is not None:
            try:
                _ = get_ome_zarr_plate(new_url)
                current_urls = st.session_state[f"{Scope.SETUP}:plate_setup:urls"]
                current_urls.add(new_url)
                st.session_state[f"{Scope.SETUP}:plate_setup:urls"] = current_urls
            except Exception as e:
                error_msg = f"Error loading plate at {new_url} \n{e}"
                st.error(error_msg)
                logger.error(error_msg)

    local_urls = st.session_state[f"{Scope.SETUP}:plate_setup:urls"]
    return local_urls


def build_plate_setup_df(plate_urls: list[str]) -> pl.DataFrame:
    plates = []
    for plate_url in plate_urls:
        plate_url = sanify_and_validate_url(plate_url)
        if plate_url is None:
            continue
        plate = get_ome_zarr_plate(plate_url)
        images_paths = asyncio.run(plate.images_paths_async())
        for path_in_plate in images_paths:
            image_url = f"{plate_url}/{path_in_plate}"
            image_url = sanify_and_validate_url(image_url)
            if image_url is None:
                continue
            extras = extras_from_url(image_url)
            plates.append(
                {
                    "image_url": image_url,
                    "plate_url": plate_url,
                    **extras,
                }
            )

    plate_setup_df = pl.DataFrame(
        plates,
        schema={
            "plate_url": pl.Utf8(),
            "plate_name": pl.Utf8(),
            "row": pl.Utf8(),
            "column": pl.Int64(),
            "path_in_well": pl.Utf8(),
            "image_url": pl.Utf8(),
        },
    )
    return plate_setup_df


# ====================================================================
#
# Plate Selection Widget:
# allow the used to select which plates to include in the analysis
#
# ====================================================================


def plate_name_selection(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Create a widget for selecting plates."""

    plate_names = plate_setup_df["plate_name"].unique().sort().to_list()
    plate_urls = plate_setup_df["plate_url"].unique().sort().to_list()

    plates = {}
    if len(plate_urls) == len(plate_names):
        for plate_url, plate_name in zip(plate_urls, plate_names, strict=True):
            plates[plate_url] = plate_name
    else:
        st.warning(
            "The plate names are not unique. The url is used to identify the plate."
        )
        logger.warning(
            "The plate names are not unique. The url is used to identify the plate."
        )
        for plate_url in plate_urls:
            plates[plate_url] = plate_url

    selected_plates_names = pills_component(
        key=f"{Scope.SETUP}:plate_setup:plate_selection",
        label="Plates",
        options=list(plates.values()),
        selection_mode="multi",
        help="Select plates to include in the analysis.",
    )

    selected_plate_urls = []
    for plate_url, plate_name in plates.items():
        if plate_name in selected_plates_names:
            selected_plate_urls.append(plate_url)
    plate_setup_df = plate_setup_df.filter(
        pl.col("plate_url").is_in(selected_plate_urls)
    )
    return plate_setup_df


# ====================================================================
#
# Load feature table
#
# ====================================================================


def _feature_table_selection_widget(
    plate_feature_tables: list[str], image_feature_tables: list[str]
) -> tuple[str, str]:
    """Create a widget for selecting the feature table."""
    image_feature_tables_suffix = " (Require Agg.)"
    image_feature_tables = [
        f"{t_name}{image_feature_tables_suffix}" for t_name in image_feature_tables
    ]
    feature_tables = plate_feature_tables + image_feature_tables

    if len(feature_tables) == 0:
        error_msg = "No feature table is common to the selected plates/images."
        st.error(error_msg)
        logger.error(error_msg)
        st.stop()

    selected_table = selectbox_component(
        key=f"{Scope.SETUP}:feature_table_selection",
        label="Select Feature Table",
        options=feature_tables,
        help="Select the feature table to join with the plate setup DataFrame.",
    )

    if image_feature_tables_suffix in selected_table:
        selected_table = selected_table.replace(image_feature_tables_suffix, "")
        mode = "image"
    else:
        mode = "plate"
    return selected_table, mode


def load_feature_table(
    plate_setup_df: pl.DataFrame,
) -> tuple[pl.DataFrame, str]:
    """Load the feature table from the plate URLs."""
    plate_feature_tables = list_plate_tables(
        plate_setup_df, filter_types="feature_table"
    )
    image_feature_tables = list_images_tables(
        plate_setup_df, filter_types="feature_table"
    )

    selected_table, mode = _feature_table_selection_widget(
        plate_feature_tables, image_feature_tables
    )

    with st.spinner("Loading feature table...", show_time=True):
        if mode == "image":
            feature_table = collect_feature_table_from_images(
                plate_setup_df, selected_table
            )
            return feature_table, selected_table

        feature_table = collect_feature_table_from_plates(
            plate_setup_df, selected_table
        )
        if feature_table is None:
            st.error(f"Feature table `{selected_table}` not found in the plate URLs.")
            logger.error(
                f"Feature table `{selected_table}` not found in the plate URLs."
            )
            st.stop()
        return feature_table, selected_table


def features_infos(feature_table: pl.DataFrame, name: str = "Feature Table"):
    """Show the first few features in the feature table."""
    st.write(
        f"Feature table: {name} correctly loaded. "
        f"Contains `{len(feature_table)}` observations and "
        f"`{len(feature_table.columns)}` features."
    )


# ====================================================================
#
# Plate Mode Setup:
# allow the user to add Zarr URLs to the DataFrame
#
# ====================================================================


def plate_mode_setup_component():
    """Setup the plate mode for the dashboard."""
    global_urls = st.session_state.get(f"{Scope.SETUP}:zarr_urls", [])

    local_urls = user_plate_url_input_component()
    logger.info(f"Local URLs: {local_urls}")
    urls = global_urls + list(local_urls)
    urls = list(set(urls))

    if not urls:
        error_msg = "No URLs provided. Please provide at least one URL."
        st.error(error_msg)
        logger.error(error_msg)
        st.stop()

    st.markdown("## Plates Selection")
    plate_setup_df = build_plate_setup_df(urls)
    plate_setup_df = plate_name_selection(plate_setup_df)

    empty_selection_warn_msg = "No plates selected. Please select at least one plate."
    if plate_setup_df.is_empty():
        st.warning(empty_selection_warn_msg)
        logger.warning(empty_selection_warn_msg)
        st.stop()

    with st.expander("Advanced Selection", expanded=False):
        images_setup = advanced_plate_selection_component(plate_setup_df)

    st.markdown("## Feature Table Selection")
    feature_table, table_name = load_feature_table(images_setup)
    features_infos(feature_table, table_name)
    return feature_table.lazy(), table_name



import numpy as np
import plotly.graph_objects as go
import polars as pl
import streamlit as st
from pydantic import BaseModel, ConfigDict

from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.utils.st_components import (
    double_slider_component,
    number_input_component,
    selectbox_component,
)
from streamlit.logger import get_logger

logger = get_logger(__name__)


class HistogramFilter(BaseModel):
    column: str
    min: float
    max: float

    model_config = ConfigDict(
        validate_assignment=True,
    )

    def apply(self, feature_frame: FeatureFrame) -> FeatureFrame:
        """
        Filter the feature frame using the histogram filter
        """
        filtered_table = feature_frame.table.filter(
            (pl.col(self.column) >= self.min) & (pl.col(self.column) <= self.max)
        )

        return FeatureFrame(
            table=filtered_table,
            features=feature_frame.features,
            cathegorical=feature_frame.cathegorical,
            others=feature_frame.others,
        )


def histogram_filter_component(
    key: str,
    feature_frame: FeatureFrame,
) -> FeatureFrame:
    """
    Create a histogram filter for the feature frame
    And return the filtered feature frame
    """
    if len(feature_frame.features) == 0:
        error_msg = "No features found in the feature table."
        logger.error(error_msg)
        raise ValueError(error_msg)

    col1, col2 = st.columns(2)
    with col1:
        column = selectbox_component(
            key=f"{key}:histogram_filter_column",
            label="Select column to filter",
            options=feature_frame.features,
        )
    with col2:
        num_bins = number_input_component(
            key=f"{key}:histogram_filter_num_bins",
            label="Number of bins",
            min_value=1,
            max_value=1000,
            value=100,
            help="Number of bins for the histogram.",
        )
    values = feature_frame.table.select(column).collect().to_series().to_numpy()
    origin_min = values.min()
    origin_max = values.max()

    min_filter, max_filter = double_slider_component(
        key=f"{key}:histogram_filter_slider",
        label="Select range to filter",
        min_value=origin_min,
        max_value=origin_max,
        help="Select the range to filter the histogram.",
    )
    filtered_values = values[np.logical_and(values >= min_filter, values <= max_filter)]
    # original

    common_bins = dict(
        start=origin_min,
        end=origin_max,
        size=(origin_max - origin_min) / num_bins,
    )

    original_histo = go.Histogram(
        x=values,
        name="Original",
        opacity=0.5,
        xbins=common_bins,
    )
    # filtered
    filtered_histo = go.Histogram(
        x=filtered_values,
        name="Filtered",
        opacity=1,
        xbins=common_bins,
        marker=dict(color="rgba(255, 127, 14, 0.5)", line=dict(color="black", width=1)),
    )
    fig = go.Figure()
    fig.add_trace(original_histo)
    fig.add_trace(filtered_histo)

    fig.update_layout(
        barmode="overlay",
        title="Original vs. Filtered Histogram",
        xaxis_title="Value",
        yaxis_title="Count",
    )
    st.plotly_chart(fig, key=f"{key}:histogram_filter_plot_overlay")
    logger.info("Histogram filter plot created")
    state = HistogramFilter(
        column=column,
        min=min_filter,
        max=max_filter,
    )

    st.session_state[f"{key}:type"] = "histogram"
    st.session_state[f"{key}:state"] = state.model_dump_json()
    feature_frame = state.apply(feature_frame)
    logger.info(f"Histogram filter applied: {state.column} [{state.min}, {state.max}]")
    return feature_frame



import streamlit as st
from streamlit.logger import get_logger
from fractal_feature_explorer.utils import Scope

from fractal_feature_explorer.authentication import verify_authentication

from fractal_feature_explorer.config import get_config

logger = get_logger(__name__)


def main():
    verify_authentication()
    current_email = st.session_state.get(f"{Scope.PRIVATE}:fractal-email", None)
    config = get_config()
    if config.deployment_type == "local":
        return None

    st.write("You are currently logged in as a Fractal user.")
    st.write(f"Fractal server URL: {config.fractal_frontend_url}")
    st.write(f"User email: {current_email}.")


if __name__ == "__main__":
    main()



import urllib3
import asyncio
from collections.abc import Iterable
from typing import Literal

import numpy as np
import streamlit as st
from ngio import (
    OmeZarrContainer,
    OmeZarrPlate,
    open_ome_zarr_container,
    open_ome_zarr_plate,
)
from ngio.common import Dimensions, Roi, list_image_tables_async
from ngio.common._zoom import numpy_zoom
from ngio.ome_zarr_meta.ngio_specs import PixelSize
from ngio.tables import MaskingRoiTable
from ngio.utils import fractal_fsspec_store

from pathlib import Path

import fsspec
from ngio.utils import NgioValueError
import urllib3.util
from fractal_feature_explorer.utils import get_fractal_token
from streamlit.logger import get_logger

from fractal_feature_explorer.config import get_config

logger = get_logger(__name__)


def _url_belongs_to_base(url: str, base_url: str) -> bool:
    """
    Check if the url has the same protocol and host as the base_url,
    and if the path of the url is a subpath of the base_url.
    """
    parsed_url = urllib3.util.parse_url(url)
    parsed_base_url = urllib3.util.parse_url(base_url)
    if (parsed_url.scheme, parsed_url.host) != (
        parsed_base_url.scheme,
        parsed_base_url.host,
    ):
        logger.debug(f"Not including token for {url=}, case 1.")
        return False
    elif parsed_url.path is not None and (
        parsed_base_url.path is None
        or not parsed_url.path.startswith(parsed_base_url.path)
    ):
        logger.debug(f"Not including token for {url=}, case 2.")
        return False
    else:
        logger.debug(f"Including token for {url=}.")
        return True


def _include_token_for_url(url: str) -> bool:
    """
    Check if the URL is a valid HTTP Fractal URL.
    """
    config = get_config()
    if config.deployment_type == "production":
        return _url_belongs_to_base(url, config.fractal_data_url)
    else:
        for data_url in config.fractal_data_urls:
            if _url_belongs_to_base(url, data_url):
                logger.debug(f"Including token for {url=}, matched with {data_url=}.")
                return True
        else:
            logger.debug(f"Not including token for {url=}, not a Fractal URL.")
            return False


def is_http_url(url: str) -> bool:
    """Check if the URL is a valid HTTP URL."""
    return url.startswith("http://") or url.startswith("https://")


@st.cache_data
def _get_http_store(
    url: str, fractal_token: str | None = None
) -> fsspec.mapping.FSMap | None:
    """Ping the URL to check if it is reachable."""
    try:
        logger.info(f"Attempting to open URL: {url}")
        store = fractal_fsspec_store(url, fractal_token=fractal_token)
    except NgioValueError as e:
        st.error(e)
        logger.error(e)
        return None
    return store


def get_http_store(
    url: str, fractal_token: str | None = None
) -> fsspec.mapping.FSMap | None:
    """Ping the URL to check if it is reachable."""
    if not _include_token_for_url(url):
        # Do not use a fractal token for non-Fractal URLs
        fractal_token = None

    return _get_http_store(url, fractal_token=fractal_token)


def get_path(url: str) -> str | None:
    """Sanitize the URL by removing the trailing slash."""
    config = get_config()
    if not config.allow_local_paths:
        st.error("Local paths are not allowed in this configuration.")
        logger.error("Local paths are not allowed in this configuration.")
        return None

    if url.startswith("~/"):
        url = url.replace("~", str(Path.home()))

    path = Path(url).resolve()
    if not path.exists():
        st.error(f"Path does not exist: {path}")
        logger.error(f"Path does not exist: {path}")
        return None
    return str(path)


def _get_and_validate_store(
    url: str, fractal_token: str | None = None
) -> fsspec.mapping.FSMap | str | None:
    """Get the store for the given URL."""
    if is_http_url(url):
        return get_http_store(url, fractal_token=fractal_token)

    return get_path(url)


def get_and_validate_store(url: str) -> fsspec.mapping.FSMap | str | None:
    """Get the store for the given URL."""
    fractal_token = get_fractal_token()
    return _get_and_validate_store(url, fractal_token=fractal_token)


@st.cache_resource
def _get_ome_zarr_plate(url: str, fractal_token: str | None = None) -> OmeZarrPlate:
    store = _get_and_validate_store(url, fractal_token=fractal_token)
    if store is None:
        raise ValueError(f"Could not get store for URL: {url}")
    plate = open_ome_zarr_plate(store, cache=True, parallel_safe=False, mode="r")
    return plate


def get_ome_zarr_plate(url: str) -> OmeZarrPlate:
    fractal_token = get_fractal_token()
    return _get_ome_zarr_plate(url, fractal_token=fractal_token)


@st.cache_resource
def _get_ome_zarr_image_container(
    url: str, fractal_token: str | None = None
) -> OmeZarrContainer:
    store = _get_and_validate_store(url, fractal_token=fractal_token)
    if store is None:
        raise ValueError(f"Could not get store for URL: {url}")
    container = open_ome_zarr_container(store, cache=True, mode="r")
    return container


def _get_ome_zarr_container_in_plate(
    url: str, fractal_token: str | None = None
) -> OmeZarrContainer:
    *_plate_url, row, col, path_in_well = url.split("/")
    plate_url = "/".join(_plate_url)
    plate = _get_ome_zarr_plate(plate_url, fractal_token=fractal_token)
    images = asyncio.run(plate.get_images_async())
    path = f"{row}/{col}/{path_in_well}"
    return images[path]


@st.cache_resource
def _get_ome_zarr_container(
    url: str,
    fractal_token: str | None = None,
    mode: Literal["image", "plate"] = "image",
):
    if mode == "plate":
        return _get_ome_zarr_container_in_plate(url, fractal_token=fractal_token)

    return _get_ome_zarr_image_container(url, fractal_token=fractal_token)


def get_ome_zarr_container(
    url: str, mode: Literal["image", "plate"] = "image"
) -> OmeZarrContainer:
    fractal_token = get_fractal_token()
    if mode == "plate":
        return _get_ome_zarr_container_in_plate(url, fractal_token=fractal_token)

    return _get_ome_zarr_image_container(url, fractal_token=fractal_token)


@st.cache_data
def _list_image_tables(
    urls: list[str],
    fractal_token: str | None = None,
    mode: Literal["image", "plate"] = "image",
) -> list[str]:
    images = [
        _get_ome_zarr_container(url, fractal_token=fractal_token, mode=mode)
        for url in urls
    ]
    image_list = asyncio.run(list_image_tables_async(images))
    return image_list


@st.cache_data
def list_image_tables(
    urls: list[str],
    fractal_token: str | None = None,
    mode: Literal["image", "plate"] = "image",
) -> list[str]:
    fractal_token = get_fractal_token()
    return _list_image_tables(
        urls=urls,
        fractal_token=fractal_token,
        mode=mode,
    )


def roi_to_slice_kwargs(
    roi: Roi,
    pixel_size: PixelSize,
    dimensions: Dimensions,
    z_slice: int = 0,
    t_slice: int = 0,
) -> dict[str, slice | int | Iterable[int]]:
    """Convert a WorldCooROI to slice_kwargs."""
    raster_roi = roi.to_pixel_roi(
        pixel_size=pixel_size, dimensions=dimensions
    ).to_slices()

    if dimensions.has_axis(axis_name="z"):
        raster_roi["z"] = z_slice  # type: ignore

    if dimensions.has_axis(axis_name="t"):
        raster_roi["t"] = t_slice  # type: ignore

    return raster_roi  # type: ignore


@st.cache_resource
def _get_masking_roi(
    image_url: str,
    ref_label: str,
    fractal_token: str | None = None,
) -> MaskingRoiTable:
    container = _get_ome_zarr_container(
        image_url,
        fractal_token=fractal_token,
        mode="image",
    )

    for table in container.list_tables(filter_types="masking_roi_table"):
        table = container.get_masking_roi_table(name=table)
        if table.reference_label == ref_label:
            table.set_table_data()
            return table

    label_img = container.get_label(name=ref_label)
    masking_roi = label_img.build_masking_roi_table()
    return masking_roi


@st.cache_data
def _get_image_array(
    image_url: str,
    ref_label: str,
    label: int,
    channel: int,
    z_slice: int = 0,
    t_slice: int = 0,
    level_path: str = "0",
    zoom_factor: float = 1,
    fractal_token: str | None = None,
) -> np.ndarray:
    container = _get_ome_zarr_container(
        image_url,
        fractal_token=fractal_token,
        mode="image",
    )
    image = container.get_image(path=level_path)
    masking_roi = _get_masking_roi(
        image_url=image_url,
        ref_label=ref_label,
        fractal_token=fractal_token,
    )
    roi = masking_roi.get(label=label)
    roi = roi.zoom(zoom_factor=zoom_factor)
    roi_slice = roi_to_slice_kwargs(
        roi=roi,
        pixel_size=image.pixel_size,
        dimensions=image.dimensions,
        z_slice=z_slice,
        t_slice=t_slice,
    )
    image_array = image.get_array(
        mode="numpy",
        c=channel,
        **roi_slice,  # type: ignore
    )
    assert isinstance(image_array, np.ndarray), "Image is not a numpy array"
    return image_array


@st.cache_data
def _get_label_array(
    image_url: str,
    ref_label: str,
    label: int,
    z_slice: int = 0,
    t_slice: int = 0,
    level_path: str = "0",
    zoom_factor: float = 1,
    fractal_token: str | None = None,
) -> np.ndarray:
    container = _get_ome_zarr_container(
        image_url,
        fractal_token=fractal_token,
        mode="image",
    )
    image = container.get_image(path=level_path)
    label_img = container.get_label(
        name=ref_label, pixel_size=image.pixel_size, strict=False
    )
    masking_roi = _get_masking_roi(
        image_url=image_url,
        ref_label=ref_label,
        fractal_token=fractal_token,
    )
    roi = masking_roi.get(label=label)
    roi = roi.zoom(zoom_factor=zoom_factor)
    roi_slice = roi_to_slice_kwargs(
        roi=roi,
        pixel_size=label_img.pixel_size,
        dimensions=label_img.dimensions,
        z_slice=z_slice,
        t_slice=t_slice,
    )
    label_array = label_img.get_array(
        mode="numpy",
        **roi_slice,  # type: ignore
    )
    assert isinstance(label_array, np.ndarray), "Label is not a numpy array"
    return label_array


def get_single_label_image(
    image_url: str,
    ref_label: str,
    label: int,
    channel: int,
    z_slice: int = 0,
    t_slice: int = 0,
    level_path: str = "0",
    show_label: bool = True,
    zoom_factor: float = 1,
) -> np.ndarray:
    """
    Get the region of interest from the image url
    """
    fractal_token = get_fractal_token()
    image_array = _get_image_array(
        image_url=image_url,
        ref_label=ref_label,
        label=label,
        channel=channel,
        z_slice=z_slice,
        t_slice=t_slice,
        level_path=level_path,
        zoom_factor=zoom_factor,
        fractal_token=fractal_token,
    )
    image_array = image_array.squeeze()
    image_array = np.clip(image_array, 0, 255)

    image_rgba = np.empty(
        (image_array.shape[0], image_array.shape[1], 4), dtype=np.uint8
    )
    image_rgba[..., 0:3] = image_array[..., np.newaxis].repeat(3, axis=2)
    image_rgba[..., 3] = 255

    if not show_label:
        return image_rgba

    label_array = _get_label_array(
        image_url=image_url,
        ref_label=ref_label,
        label=label,
        z_slice=z_slice,
        t_slice=t_slice,
        level_path=level_path,
        zoom_factor=zoom_factor,
        fractal_token=fractal_token,
    )

    label_array = label_array.squeeze()
    label_array = np.where(label_array == label, 255, 0)

    # Scale the label array to match the image size
    label_array = numpy_zoom(
        label_array,
        target_shape=image_rgba.shape[:2],
        order=0,  # Always use nearest neighbor for labels
    )

    image_rgba[label_array > 0, 0] = 255

    return image_rgba



import copy

import plotly.express as px
import streamlit as st

from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.utils.st_components import (
    selectbox_component,
)


def heat_map_component(
    key: str,
    feature_frame: FeatureFrame,
) -> None:
    features_columns = feature_frame.features
    selected_feature = selectbox_component(
        key=f"{key}:scatter_plot_feature",
        label="**Feature**",
        options=features_columns,
    )

    cathegorical_columns = copy.deepcopy(feature_frame.cathegorical)

    if "column" in cathegorical_columns:
        cathegorical_columns.remove("column")
        cathegorical_columns = ["column"] + cathegorical_columns

    x_axis = selectbox_component(
        key=f"{key}:scatter_plot_x_axis",
        label="**X-axis**",
        options=cathegorical_columns,
    )

    x_axis_index = cathegorical_columns.index(x_axis)
    cathegorical_columns.pop(x_axis_index)

    if "row" in cathegorical_columns:
        cathegorical_columns.remove("row")
        cathegorical_columns = ["row"] + cathegorical_columns

    y_axis = selectbox_component(
        key=f"{key}:scatter_plot_y_axis",
        label="**Y-axis**",
        options=cathegorical_columns,
    )

    y_axis_index = cathegorical_columns.index(y_axis)
    cathegorical_columns.pop(y_axis_index)

    aggregation = st.pills(
        label="**Aggregation**",
        options=["Mean", "Sum", "Median", "Counts"],
        default="Mean",
        key=f"{key}:scatter_plot_aggregation",
        selection_mode="single",
        help="Select the type of aggregation to apply.",
    )
    axes_names = [x_axis, y_axis]
    columns_needed = [selected_feature] + axes_names
    feature_df = feature_frame.table.select(columns_needed).collect()
    feature_df = feature_df.to_pandas()

    df_piv = feature_df.groupby(axes_names, as_index=False)
    if aggregation == "Mean":
        df_piv = df_piv.mean(numeric_only=True)
    elif aggregation == "Sum":
        df_piv = df_piv.sum(numeric_only=True)
    elif aggregation == "Median":
        df_piv = df_piv.median(numeric_only=True)
    elif aggregation == "Counts":
        df_piv = df_piv.count()
    else:
        raise ValueError(f"Unknown aggregation: {aggregation}")

    new_feature_name = f"{selected_feature} {aggregation}"
    df_piv = df_piv.rename(columns={selected_feature: new_feature_name})
    df_piv = df_piv.pivot(index=x_axis, columns=y_axis, values=new_feature_name)
    img = df_piv.to_numpy()
    fig = px.imshow(
        img.T,
        x=df_piv.index.to_list(),
        y=df_piv.columns.to_list(),
        labels={"x": x_axis, "y": y_axis, "color": new_feature_name},
    )

    fig.update_xaxes(  # type: ignore
        type="category", showgrid=True, title=x_axis, tickson="boundaries", ticklen=0
    )
    fig.update_yaxes(  # type: ignore
        type="category", showgrid=True, title=y_axis, tickson="boundaries", ticklen=0
    )
    st.plotly_chart(fig)



"""
Fractal Feature Explorer - Setup Page
"""

import streamlit as st
from streamlit.logger import get_logger

from fractal_feature_explorer.pages.setup_page._plate_mode_setup import (
    plate_mode_setup_component,
)
from fractal_feature_explorer.utils import Scope, invalidate_session_state
import polars as pl
from fractal_feature_explorer.config import get_config
from fractal_feature_explorer.authentication import verify_authentication

logger = get_logger(__name__)


def init_global_state():
    if f"{Scope.PRIVATE}:fractal-token" not in st.session_state:
        st.session_state[f"{Scope.PRIVATE}:fractal-token"] = None
    if f"{Scope.SETUP}:setup_mode" not in st.session_state:
        st.session_state[f"{Scope.SETUP}:setup_mode"] = "Plates"
    if f"{Scope.SETUP}:zarr_urls" not in st.session_state:
        st.session_state[f"{Scope.SETUP}:zarr_urls"] = []


def parse_query_params():
    setup_mode = st.query_params.get("setup_mode", None)
    if setup_mode is not None:
        st.session_state[f"{Scope.SETUP}:setup_mode"] = setup_mode
        logger.info(f"setup_mode: {setup_mode} (set url from query params)")

    zarr_urls = st.query_params.get_all("zarr_url")
    if len(zarr_urls) > 0:
        _zarr_urls = st.session_state.get(f"{Scope.SETUP}:zarr_urls", [])
        st.session_state[f"{Scope.SETUP}:zarr_urls"] = _zarr_urls + zarr_urls
        logger.info(f"zarr_urls: {zarr_urls} (set url from query params)")


def setup_global_state():
    """
    Setup the global state for the Streamlit app.
    """
    init_global_state()
    parse_query_params()  # This may be useful e.g. when linking from fractal-web

    default_setup_mode = st.session_state.get(f"{Scope.SETUP}:setup_mode", "Plates")
    setup_mode = st.pills(
        "Setup Mode",
        options=["Plates", "Images"],
        default=default_setup_mode,
        key=f"{Scope.SETUP}:_setup_mode",
        help="Select the mode of the setup.",
    )
    st.session_state[f"{Scope.SETUP}:setup_mode"] = setup_mode
    return setup_mode


def filter_cache_invalidations(
    features_table: pl.LazyFrame, table_name: str
) -> pl.Schema:
    schema = features_table.collect_schema()
    if f"{Scope.DATA}:feature_table" in st.session_state:
        old_table_name = st.session_state[f"{Scope.DATA}:feature_table_name"]
        old_schema = st.session_state[f"{Scope.DATA}:feature_table_schema"]
        if old_table_name != table_name:
            # invalidate the old table
            warn_msg = (
                f"The feature table name has changed. {old_table_name} -> {table_name}. \n"
                "All filters have been reset."
            )
            logger.warning(warn_msg)
            st.warning(warn_msg)
            invalidate_session_state(f"{Scope.FILTERS}")

        elif old_schema != schema:
            # invalidate the old table
            warn_msg = (
                "The feature table schema has changed. The filters have been reset."
            )
            logger.warning(warn_msg)
            st.warning(warn_msg)
            invalidate_session_state(f"{Scope.FILTERS}")
    return schema


def _token_input_widget():
    """
    Input field for the Fractal authentication token.
    """
    config = get_config()
    if config.deployment_type == "production":
        return None
    current_token = st.session_state.get(f"{Scope.PRIVATE}:fractal-token", None)
    current_token = current_token if current_token else ""
    token = st.text_input(
        label="Fractal Authentication Token",
        value=current_token,
        key="_fractal_token",
        type="password",
    )
    if token == "":
        st.session_state[f"{Scope.PRIVATE}:fractal-token"] = None
    else:
        st.session_state[f"{Scope.PRIVATE}:fractal-token"] = token


def main():
    verify_authentication()

    setup_mode = setup_global_state()

    with st.sidebar:
        with st.expander("Advanced Options", expanded=False):
            _token_input_widget()

            st.divider()
            if st.button(
                "Reset Setup",
                key=f"{Scope.SETUP}:reset_setup",
                icon="🔄",
                help="Reset the setup state. This will clear all filters and the feature table.",
            ):
                invalidate_session_state(f"{Scope.SETUP}")
                st.rerun()

    match setup_mode:
        case "Plates":
            features_table, table_name = plate_mode_setup_component()
        case "Images":
            st.error("Image mode is not yet implemented. Please select 'Plates' mode.")
            logger.error("Image mode is not yet implemented.")
            st.stop()
        case _:
            error_msg = f"Invalid setup mode selected. Should be 'Plates' or 'Images' but got {setup_mode}."
            st.error(error_msg)
            logger.error(error_msg)
            st.stop()

    schema = filter_cache_invalidations(features_table, table_name)

    st.session_state[f"{Scope.DATA}:feature_table"] = features_table
    st.session_state[f"{Scope.DATA}:feature_table_name"] = table_name
    st.session_state[f"{Scope.DATA}:feature_table_schema"] = schema
    logger.info(
        f"Feature table {table_name} with schema {schema} has been set in session state."
    )
    logger.info("Setup page loading complete.")


if __name__ == "__main__":
    main()



import polars as pl
import streamlit as st

from fractal_feature_explorer.utils.common import Scope
from fractal_feature_explorer.utils.st_components import (
    double_slider_component,
    multiselect_component,
    pills_component,
    selectbox_component,
)

from fractal_feature_explorer.pages.setup_page._tables_io import (
    list_images_tables,
    list_plate_tables,
    collect_condition_table_from_images,
    collect_condition_table_from_plates,
)

from streamlit.logger import get_logger

logger = get_logger(__name__)


def plate_name_from_url(plate_url: str) -> str:
    """Get the plate name from the URL."""
    return plate_url.rsplit("/", 1)[-1]


def extras_from_url(image_url: str) -> dict[str, str]:
    """Get the extras from the URL."""
    *_, plate_name, row, column, path_in_well = image_url.split("/")
    return {
        "plate_name": plate_name,
        "row": row,
        "column": column,
        "path_in_well": path_in_well,
    }


# ====================================================================
#
# Row Selection Widget:
# allow the used to select which rows to include in the analysis
#
# ====================================================================


def rows_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Create a widget for selecting rows."""
    rows = plate_setup_df["row"].unique().sort().to_list()
    rows_names = pills_component(
        key=f"{Scope.SETUP}:plate_setup:row_selection",
        label="Rows",
        options=rows,
        selection_mode="multi",
        help="Select rows to include in the analysis.",
    )

    plate_setup_df = plate_setup_df.filter(pl.col("row").is_in(rows_names))
    return plate_setup_df


# ====================================================================
#
# Column Selection Widget:
# allow the used to select which columns to include in the analysis
#
# ====================================================================


def _columns_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> list[str]:
    """Create a widget for selecting columns."""
    columns = plate_setup_df["column"].unique().sort().to_list()
    columns_names = pills_component(
        key=f"{Scope.SETUP}:plate_setup:columns_selection",
        label="Columns",
        options=columns,
        selection_mode="multi",
        help="Select columns to include in the analysis.",
    )
    return columns_names


def _columns_selection_filter(
    plate_setup_df: pl.DataFrame, columns_names: list[str]
) -> pl.DataFrame:
    """Filter the plate setup DataFrame based on the selected columns."""
    plate_setup_df = plate_setup_df.filter(pl.col("column").is_in(columns_names))
    return plate_setup_df


def columns_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Create a widget for selecting columns."""
    columns_names = _columns_selection_widget(plate_setup_df)
    plate_setup_df = _columns_selection_filter(plate_setup_df, columns_names)
    return plate_setup_df


# ====================================================================
#
# Acquisition Selection Widget:
# allow the used to select which acquisitions to include in the analysis
#
# ====================================================================


def _acquisition_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> list[str]:
    """Create a widget for selecting acquisitions."""
    acquisitions = plate_setup_df["path_in_well"].unique().sort().to_list()
    acquisitions_names = pills_component(
        key=f"{Scope.SETUP}:plate_setup:acquisition_selection",
        label="Acquisitions",
        options=acquisitions,
        selection_mode="multi",
        help="Select acquisitions to include in the analysis.",
    )
    return acquisitions_names


def _acquisition_selection_filter(
    plate_setup_df: pl.DataFrame, acquisitions_names: list[str]
) -> pl.DataFrame:
    """Filter the plate setup DataFrame based on the selected acquisitions."""
    plate_setup_df = plate_setup_df.filter(
        pl.col("path_in_well").is_in(acquisitions_names)
    )
    return plate_setup_df


def acquisition_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Create a widget for selecting acquisitions."""
    acquisitions_names = _acquisition_selection_widget(plate_setup_df)
    plate_setup_df = _acquisition_selection_filter(plate_setup_df, acquisitions_names)
    return plate_setup_df


# ====================================================================
#
# Wells Selection Widget:
# allow the used to select which wells to include in the analysis
#
# ====================================================================


def _wells_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> list[str]:
    """Create a widget for selecting wells."""
    wells = plate_setup_df["well_id"].unique().sort().to_list()
    wells_names = pills_component(
        key=f"{Scope.SETUP}:plate_setup:wells_selection",
        label="Wells",
        options=wells,
        selection_mode="multi",
        help="Select wells to include in the analysis.",
    )
    return wells_names


def _wells_selection_filter(
    plate_setup_df: pl.DataFrame, wells_names: list[str]
) -> pl.DataFrame:
    """Filter the plate setup DataFrame based on the selected wells."""
    plate_setup_df = plate_setup_df.filter(pl.col("well_id").is_in(wells_names))
    return plate_setup_df


def wells_selection_widget(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Create a widget for selecting wells."""
    plate_setup_df = plate_setup_df.with_columns(
        pl.concat_str(["row", "column"], separator="").alias("well_id"),
    )
    wells_names = _wells_selection_widget(plate_setup_df)
    plate_setup_df = _wells_selection_filter(plate_setup_df, wells_names)
    plate_setup_df = plate_setup_df.drop("well_id")
    return plate_setup_df


# ====================================================================
#
# Find condition tables and join them with the plate setup DataFrame
#
# ====================================================================


def _join_condition_table_widget(
    table_condition_tables: list[str], image_condition_tables: list[str]
) -> tuple[str | None, str | None]:
    """Create a widget for selecting the condition table."""
    image_condition_tables_suffix = " (Require Agg.)"
    image_condition_tables = [
        f"{t_name}{image_condition_tables_suffix}" for t_name in image_condition_tables
    ]
    condition_tables = (
        ["-- No Condition Table --"] + table_condition_tables + image_condition_tables
    )
    selected_table = selectbox_component(
        key=f"{Scope.SETUP}:plate_setup:condition_table_selection",
        label="Select Condition Table",
        options=condition_tables,
    )
    if selected_table == "-- No Condition Table --":
        return None, None

    if image_condition_tables_suffix in selected_table:
        selected_table = selected_table.replace(image_condition_tables_suffix, "")
        mode = "image"
    else:
        mode = "plate"
    return selected_table, mode


def join_condition_tables(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Join the condition table with the plate setup DataFrame."""
    plate_condition_tables = list_plate_tables(
        plate_setup_df, filter_types="condition_table"
    )
    image_condition_tables = list_images_tables(
        plate_setup_df, filter_types="condition_table"
    )
    selected_table, mode = _join_condition_table_widget(
        plate_condition_tables, image_condition_tables
    )
    if selected_table is None:
        return plate_setup_df

    if mode == "plate":
        new_plate_setup_df = collect_condition_table_from_plates(
            plate_setup_df=plate_setup_df,
            table_name=selected_table,
        )
        if new_plate_setup_df is None:
            st.warning(f"Condition table {selected_table} not found in the plates.")
            return plate_setup_df
        return new_plate_setup_df

    return collect_condition_table_from_images(plate_setup_df, selected_table)


# ====================================================================
#
# Filter Based on Condition Tables
# allow the user to filter the images based on the condition tables
#
# ====================================================================


def _numeric_filter_widget(plate_setup_df: pl.DataFrame, column: str) -> pl.DataFrame:
    """Create a widget for filtering numeric columns."""
    series = plate_setup_df[column]

    min_value = series.min()
    max_value = series.max()

    if min_value == max_value:
        return plate_setup_df

    # ignore the type error
    # this is a numeric column

    selected_range = double_slider_component(
        key=f"{Scope.SETUP}:plate_setup:numeric_filter_{column}",
        label=f"Filter {column}:",
        min_value=min_value,  # type: ignore
        max_value=max_value,  # type: ignore
        help=f"Select range to include in the analysis for the column {column}.",
    )

    plate_setup_df = plate_setup_df.filter(
        (pl.col(column) >= selected_range[0]) & (pl.col(column) <= selected_range[1])
    )
    return plate_setup_df


def _boolean_filter_widget(plate_setup_df: pl.DataFrame, column: str) -> pl.DataFrame:
    """Create a widget for filtering boolean columns."""
    selected_value = pills_component(
        key=f"{Scope.SETUP}:plate_setup:boolean_filter_{column}",
        label=f"Filter {column}:",
        options=[True, False],
        selection_mode="multi",
        help=f"Select values to include in the analysis for the column {column}.",
    )
    plate_setup_df = plate_setup_df.filter(pl.col(column).is_in(selected_value))
    return plate_setup_df


def _string_filter_widget(plate_setup_df: pl.DataFrame, column: str) -> pl.DataFrame:
    """Create a widget for filtering string columns."""
    unique_values = plate_setup_df[column].unique().sort().to_list()
    selected_values = multiselect_component(
        key=f"{Scope.SETUP}:plate_setup:string_filter_{column}",
        label=f"Filter {column}:",
        options=unique_values,
        help=f"Select values to include in the analysis for the column {column}.",
    )
    plate_setup_df = plate_setup_df.filter(pl.col(column).is_in(selected_values))
    return plate_setup_df


def filter_based_on_condition(plate_setup_df: pl.DataFrame) -> pl.DataFrame:
    condition_columns = set(plate_setup_df.columns) - set(
        [
            "plate_url",
            "plate_name",
            "row",
            "column",
            "path_in_well",
            "image_url",
        ]
    )

    for column in condition_columns:
        column_series: pl.Series = plate_setup_df[column]
        column_type = column_series.dtype

        if column_type.is_numeric():
            plate_setup_df = _numeric_filter_widget(plate_setup_df, column)
        elif column_type == pl.Boolean:
            plate_setup_df = _boolean_filter_widget(plate_setup_df, column)
        elif column_type == pl.String():
            plate_setup_df = _string_filter_widget(plate_setup_df, column)
        else:
            st.warning(
                f"Column {column} is of type {column_type}. Filtering not supported."
            )

    return plate_setup_df


def into_images_df(plate_setup_df: pl.DataFrame) -> pl.DataFrame:
    """Convert the plate setup DataFrame into a DataFrame of images."""
    plate_setup_df = plate_setup_df.group_by(
        ["plate_url", "row", "column", "path_in_well"]
    ).all()
    for col in plate_setup_df.columns:
        df_c = plate_setup_df[col]
        if df_c.dtype == pl.List and df_c.list.unique().list.len().eq(1).all():
            plate_setup_df = plate_setup_df.with_columns(
                pl.col(col).list.first().alias(col),
            )
    return plate_setup_df


def show_selected_images_widget(images_df: pl.DataFrame):
    """Show the selected images in the plate setup DataFrame."""
    images_df = images_df.drop("plate_url")
    st.dataframe(images_df)
    st.write("Images Selected: ", len(images_df))


def advanced_plate_selection_component(
    plate_setup_df: pl.DataFrame,
) -> pl.DataFrame:
    """Advanced selection of images from a plate."""
    empty_selection_warn_msg = "No plates selected. Please select at least one plate."
    col1, col2 = st.columns(2)
    with col1:
        plate_setup_df = rows_selection_widget(plate_setup_df)
    with col2:
        plate_setup_df = columns_selection_widget(plate_setup_df)
    plate_setup_df = acquisition_selection_widget(plate_setup_df)

    if st.toggle(
        key=f"{Scope.SETUP}:plate_setup:toggle_wells_selection",
        label="Select Specific Wells",
        value=False,
    ):
        plate_setup_df = wells_selection_widget(plate_setup_df)

    if plate_setup_df.is_empty():
        st.warning(empty_selection_warn_msg)
        logger.warning(empty_selection_warn_msg)
        st.stop()

    if st.toggle(
        key=f"{Scope.SETUP}:plate_setup:toggle_condition_table",
        label="Load Condition Table",
        value=False,
    ):
        st.markdown("## Condition Tables")
        with st.spinner("Loading condition tables...", show_time=True):
            plate_setup_df = join_condition_tables(plate_setup_df)
            plate_setup_df = filter_based_on_condition(plate_setup_df)

        if plate_setup_df.is_empty():
            st.warning(empty_selection_warn_msg)
            logger.warning(empty_selection_warn_msg)
            st.stop()

    st.markdown("## Final Images Selection")
    images_setup = into_images_df(plate_setup_df)
    show_selected_images_widget(images_setup)

    return images_setup



from dataclasses import dataclass, field

import polars as pl


@dataclass
class FeatureFrame:
    table: pl.LazyFrame
    features: list[str] = field(default_factory=list)
    cathegorical: list[str] = field(default_factory=list)
    others: list[str] = field(default_factory=list)

    @property
    def protected(self) -> list[str]:
        return ["image_url", "reference_label", "label"]

    def all_columns(self) -> list[str]:
        """
        Get all columns in the feature table.
        """
        return self.features + self.cathegorical + self.others



from fractal_feature_explorer.utils.common import (
    Scope,
    invalidate_session_state,
    get_fractal_token,
)
from fractal_feature_explorer.config import get_config
from fractal_feature_explorer.utils.st_components import (
    double_slider_component,
    multiselect_component,
    number_input_component,
    pills_component,
    selectbox_component,
    single_slider_component,
)

from fractal_feature_explorer.utils.ngio_io_caches import (
    get_ome_zarr_plate,
    get_ome_zarr_container,
    get_and_validate_store,
    get_single_label_image,
    list_image_tables,
    is_http_url,
)

__all__ = [
    "get_config",
    "pills_component",
    "selectbox_component",
    "multiselect_component",
    "double_slider_component",
    "single_slider_component",
    "number_input_component",
    "Scope",
    "invalidate_session_state",
    "get_fractal_token",
    "get_ome_zarr_plate",
    "get_ome_zarr_container",
    "get_and_validate_store",
    "is_http_url",
    "get_single_label_image",
    "list_image_tables",
]



from enum import StrEnum

import streamlit as st
from streamlit.logger import get_logger


logger = get_logger(__name__)


class Scope(StrEnum):
    """
    Enum for the different phases of the fractal explorer.
    """

    PRIVATE = "private"  # User private data (like auth token)
    SETUP = "setup"  # Setup page keys
    FILTERS = "filters"  # Filters page keys
    EXPLORE = "explore"  # Explore page keys
    DATA = "data"  # Non-serializable data (like polars DataFrames)


def invalidate_session_state(key_prefix: str) -> None:
    """
    Invalidate the session state for the given key.
    """
    logger.info(f"Invalidating session state for key prefix: {key_prefix}")
    for key in st.session_state.keys():
        _key = str(key)
        if _key.startswith(key_prefix):
            logger.info(f"Deleting key: {key}")
            del st.session_state[key]


def get_fractal_token() -> str | None:
    """
    Get the Fractal token from the session state.
    """
    return st.session_state.get(f"{Scope.PRIVATE}:fractal-token", None)



"""CLI for the Fractal Feature Explorer."""

import sys
import runpy

from pathlib import Path

CLI_ARGS = [
    "--theme.primaryColor=#0099ff",
    "--theme.backgroundColor=#FFFFFF",
    "--theme.secondaryBackgroundColor=#F0F2F6",
    "--theme.textColor=#31333F",
    "--logger.level=info",
    '--logger.messageFormat="%(asctime)s - %(levelname)s - [%(name)s:%(lineno)d] - %(message)s"',
    #'--server.headless=true',
    "--server.runOnSave=false",
    #'--server.enableCORS=true',
    #'--server.enableXsrfProtection=true',
    "--server.fileWatcherType=none",
    "--browser.gatherUsageStats=false",
    "--client.toolbarMode=minimal",
]


def cli():
    """Run the Fractal Feature Explorer CLI."""
    main_file = Path(__file__).parent / "main.py"
    sys.argv = ["streamlit", "run", str(main_file)] + CLI_ARGS
    runpy.run_module("streamlit", run_name="__main__")


if __name__ == "__main__":
    cli()



import polars as pl
import streamlit as st

from fractal_feature_explorer.pages.filters_page._column_filter import (
    ColumnsFilter,
    columns_filter_component,
)
from fractal_feature_explorer.pages.filters_page._common import FeatureFrame
from fractal_feature_explorer.pages.filters_page._histogram_filter import (
    HistogramFilter,
    histogram_filter_component,
)
from fractal_feature_explorer.pages.filters_page._scatter_filter import (
    ScatterFilter,
    scatter_filter_component,
)
from fractal_feature_explorer.utils import Scope, invalidate_session_state
from streamlit.logger import get_logger
from fractal_feature_explorer.authentication import verify_authentication

logger = get_logger(__name__)


def build_feature_frame(feature_table: pl.LazyFrame) -> FeatureFrame:
    schema = feature_table.collect_schema()

    feature = []
    cathegorical = []
    others = []
    for name, dtype in schema.items():
        if name in [
            "image_url",
            "reference_label",
            "label",
            "row",
            "column",
            "path_in_well",
            "plate_name",
            "image_name",
        ]:
            cathegorical.append(name)
        elif dtype == pl.UInt8() or dtype == pl.String() or dtype == pl.Categorical():
            cathegorical.append(name)
        elif dtype == pl.Boolean():
            cathegorical.append(name)
        elif dtype.is_numeric():
            feature.append(name)
        else:
            others.append(name)

    return FeatureFrame(
        table=feature_table,
        features=feature,
        cathegorical=cathegorical,
        others=others,
    )


def _find_unique_name(keys: list[str], prefix: str) -> str:
    """
    Find a unique name for the filter
    """
    i = 1
    while True:
        name = f"{prefix} {i}"
        if name not in keys:
            return name
        i += 1


def add_filters() -> None:
    """
    Dynamically add filters to the feature table
    """
    if f"{Scope.FILTERS}:filters_dict" not in st.session_state:
        st.session_state[f"{Scope.FILTERS}:filters_dict"] = {
            "Columns Filter": (
                f"{Scope.FILTERS}:columns_filter",
                columns_filter_component,
            ),
        }

    filter_type = st.pills(
        label="Filter Type",
        options=["Histogram Filter", "Scatter Filter"],
        default="Histogram Filter",
        key=f"{Scope.FILTERS}:filter_type",
        selection_mode="single",
        help="Select the type of filter to add.",
    )

    if st.button("Add New Filter", key=f"{Scope.FILTERS}:add_filter_button"):
        if filter_type == "Histogram Filter":
            name = _find_unique_name(
                st.session_state[f"{Scope.FILTERS}:filters_dict"].keys(),
                "Histogram Filter",
            )
            key = f"{Scope.FILTERS}:{name}_histogram_filter"
            st.session_state[f"{Scope.FILTERS}:filters_dict"][name] = (
                key,
                histogram_filter_component,
            )
            logger.info(f"New Histogram Filter added: {name}")
            st.rerun()
        elif filter_type == "Scatter Filter":
            name = _find_unique_name(
                st.session_state[f"{Scope.FILTERS}:filters_dict"].keys(),
                "Scatter Filter",
            )
            key = f"{Scope.FILTERS}:{name}_scatter_filter"
            st.session_state[f"{Scope.FILTERS}:filters_dict"][name] = (
                key,
                scatter_filter_component,
            )
            logger.info(f"New Scatter Filter added: {name}")
            st.rerun()

    return None


def display_filters(feature_frame: FeatureFrame) -> FeatureFrame:
    """
    Display the filters in the feature table
    """
    filter_list = st.session_state[f"{Scope.FILTERS}:filters_dict"]

    for name, (filter_key, filter_component) in filter_list.items():
        if "Columns Filter" in name:
            expanded = False
        else:
            expanded = True
        with st.expander(f"{name}", expanded=expanded):
            st.markdown(
                f"""
                ### {name}
                """
            )
            try:
                feature_frame = filter_component(
                    key=filter_key,
                    feature_frame=feature_frame,
                )
            except Exception as e:
                error_msg = (
                    f"Error applying filter {name}: {e}. "
                    "Please check the filter parameters and try again."
                )
                st.error(error_msg)
                logger.error(error_msg)

            col1, col2 = st.columns(2)
            with col1:
                if st.button(
                    "Reset Filter", key=f"{filter_key}:reset_filter_button", icon="🔄"
                ):
                    invalidate_session_state(filter_key)
                    st.rerun()
            with col2:
                if "Columns Filter" in name:
                    continue
                if st.button(
                    "Delete Filter", key=f"{filter_key}:delete_filter_button", icon="🚮"
                ):
                    invalidate_session_state(filter_key)
                    del filter_list[name]
                    st.session_state[f"{Scope.FILTERS}:filters_dict"] = filter_list
                    st.rerun()

    return feature_frame


def apply_filters(feature_frame: FeatureFrame) -> FeatureFrame:
    """
    Apply the filters to the feature table
    """
    if f"{Scope.FILTERS}:filters_dict" not in st.session_state:
        return feature_frame
    filters_dict = st.session_state[f"{Scope.FILTERS}:filters_dict"]

    for name, (filter_key, filter_component) in filters_dict.items():
        status_key = f"{filter_key}:state"
        type_key = f"{filter_key}:type"

        status_json = st.session_state.get(status_key, None)
        if status_json is None:
            warn_msg = (
                f"Filter {name} has not been applied yet. "
                "Please apply the filter from the filter page first."
            )
            logger.warning(warn_msg)
            st.warning(warn_msg)
            continue

        filter_type = st.session_state.get(type_key, None)
        if filter_type is None:
            warn_msg = (
                f"Filter {name} has not been applied yet. "
                "Please apply the filter from the filter page first."
            )
            logger.warning(warn_msg)
            st.warning(warn_msg)
            continue

        logger.info(f"Applying filter {name} of type {filter_type}")
        if filter_type == "columns":
            filter_component = ColumnsFilter.model_validate_json(status_json)
            feature_frame = filter_component.apply(feature_frame)
        elif filter_type == "histogram":
            filter_component = HistogramFilter.model_validate_json(status_json)
            feature_frame = filter_component.apply(feature_frame)
        elif filter_type == "scatter":
            filter_component = ScatterFilter.model_validate_json(status_json)
            feature_frame = filter_component.apply(feature_frame)
        else:
            st.warning(f"Filter {name} is not found. Please apply the filter first.")
            continue

    logger.info("Filters applied to feature table")
    return feature_frame


def feature_filters_manger(
    feature_table: pl.LazyFrame, table_name: str
) -> FeatureFrame:
    """
    Setup the feature table for the dashboard.
    """
    st.markdown(
        f"""
        ## Feature Table Setup
        Select the features to include in the feature table.
        
        Table Name: **{table_name}**
        """
    )
    feature_frame = build_feature_frame(feature_table)

    col1, _ = st.columns(2)
    with col1:
        add_filters()

    feature_frame = display_filters(feature_frame)
    return feature_frame


def main():
    verify_authentication()
    with st.sidebar:
        with st.expander("Advanced Options", expanded=False):
            if st.button(
                "Reset Filters",
                key=f"{Scope.FILTERS}:reset_filters",
                icon="🔄",
                help="Reset the filters state. This will clear all filters.",
            ):
                invalidate_session_state(f"{Scope.FILTERS}")
                st.rerun()

    feature_table = st.session_state.get(f"{Scope.DATA}:feature_table", None)
    feature_table_name = st.session_state.get(f"{Scope.DATA}:feature_table_name", "")
    if feature_table is None:
        st.warning(
            "No feature table found in session state. Please make sure to run the setup page first."
        )
        st.stop()

    feature_filters_manger(feature_table=feature_table, table_name=feature_table_name)
    logger.info("Filters page loading complete")


if __name__ == "__main__":
    main()



"""Testing Placeholder for the Fractal Feature Explorer App.

TODO This is just a placeholder for the app testing. TBD what to test.

"""

from streamlit.testing.v1 import AppTest


def test_app():
    """
    Basic Workflow Test for the Fractal Feature Explorer App.
    """
    app = AppTest.from_file("src/fractal_feature_explorer/main.py")
    app.run(timeout=15)

    assert not app.exception, "App raised an exception during execution."



from pathlib import Path

import streamlit as st


@st.dialog("File Picker", width="large")
def _file_path_picker(fpp_base_path: str | None = None) -> None:
    st.title("Simple File Browser")

    # Let the user specify a base directory (defaults to current working directory)
    def on_fpp_base_path_change():
        st.session_state.fpp_base_path = st.session_state.fpp_path_select

    def format_func(path: str | Path) -> str:
        path = str(path)
        if len(path) > 80:
            path = path[-80:]
            return f"..{path}"
        return path

    if fpp_base_path is None:
        fpp_base_path = st.session_state.get("fpp_base_path", str(Path.home()))

    if not isinstance(fpp_base_path, str):
        fpp_base_path = str(fpp_base_path)

    if not Path(fpp_base_path).exists():
        st.error(f"Base directory does not exist: {fpp_base_path}")
        st.rerun(scope="app")

    st.session_state.fpp_base_path = fpp_base_path

    st.write("Pick a file or directory:")
    all_files = [str(f) for f in Path(fpp_base_path).glob("*")]
    all_files = [Path(fpp_base_path), *all_files]

    selected_path = st.selectbox(
        "Select a directory",
        options=all_files,
        index=0,
        format_func=format_func,
        on_change=on_fpp_base_path_change,
        key="fpp_path_select",
    )

    col1, col2, col3, _ = st.columns(4)
    with col1:
        if st.button("Go back"):
            st.session_state.fpp_base_path = str(
                Path(st.session_state.fpp_base_path).parent
            )
            st.rerun(scope="fragment")

    with col2:
        if st.button("Home"):
            st.session_state.fpp_base_path = str(Path.home())
            st.rerun(scope="fragment")

    with col3:
        if st.button("Pick Current Selection"):
            st.session_state.fpp_selected_path = str(selected_path)
            st.rerun(scope="app")

    st.stop()


def file_path_picker(fpp_base_path: str | None = None) -> str | None:
    if st.button("Pick a file"):
        _file_path_picker(fpp_base_path=fpp_base_path)
    file_path = st.session_state.get("fpp_selected_path", None)
    return file_path



import asyncio

import polars as pl
import streamlit as st
from ngio.common import (
    concatenate_image_tables_async,
    concatenate_image_tables_as_async,
)
from ngio.tables import FeatureTable
from typing import Literal
from fractal_feature_explorer.utils import (
    get_ome_zarr_container,
    get_ome_zarr_plate,
)
from fractal_feature_explorer.pages.setup_page._utils import (
    plate_name_from_url,
    extras_from_url,
)
from ngio.common import list_image_tables_async

from streamlit.logger import get_logger

logger = get_logger(__name__)


def list_plate_tables(
    plate_setup_df: pl.DataFrame,
    filter_types: str = "condition_table",
    mode: Literal["all", "common"] = "common",
) -> list[str]:
    """Collect existing tables from the plate URLs."""
    plate_urls = plate_setup_df["plate_url"].unique().to_list()
    plate_tables = {}

    for url in plate_urls:
        logger.warning(f"Loading plate {url}...")

        plate = get_ome_zarr_plate(url)
        if plate._tables_container is None:
            logger.warning(
                f"Plate {url} does not contain any Plate level tables. Skipping."
            )
            continue
        try:
            list_tables = plate.list_tables(filter_types=filter_types)
            for table_name in list_tables:
                if table_name not in plate_tables:
                    plate_tables[table_name] = []
                plate_tables[table_name].append(url)
            logger.info(f"List of plate tables in {url}: {list_tables}")
        except Exception as e:
            erro_msg = f"Error loading {filter_types} tables from {url}. "
            st.error(erro_msg)
            logger.error(erro_msg)
            raise e

    if mode == "all":
        logger.info(f"List of plate level tables: {plate_tables}")
        return list(plate_tables.keys())
    elif mode == "common":
        common_tables = []
        logger.info(f"List of common plate level tables: {plate_tables}")
        for table_name, urls in plate_tables.items():
            if len(urls) == len(plate_urls):
                common_tables.append(table_name)
        return common_tables
    else:
        raise ValueError(f"Invalid mode {mode}. Must be 'all' or 'common'.")


def list_images_tables(
    plate_setup_df: pl.DataFrame,
    filter_types: str = "condition_table",
    mode: Literal["all", "common"] = "common",
) -> list[str]:
    """Collect existing image tables from the plate URLs."""
    images_urls = plate_setup_df["image_url"].unique().to_list()
    images = [get_ome_zarr_container(url, mode="plate") for url in images_urls]
    images_condition_tables = asyncio.run(
        list_image_tables_async(images=images, filter_types=filter_types, mode=mode)
    )
    if mode == "all":
        logger.info(f"List of image level tables: {images_condition_tables}")
    elif mode == "common":
        logger.info(f"List of common image level tables: {images_condition_tables}")
    return images_condition_tables


# ====================================================================
#
# Condition tables utils
#
# ====================================================================


@st.cache_data
def _load_single_plate_condition_table(
    url: str,
    table_name: str,
) -> pl.DataFrame | None:
    """Load the condition table from a single plate URL."""
    plate = get_ome_zarr_plate(url)
    try:
        table = plate.get_table(table_name)
    except Exception as e:
        st.warning(f"Error loading condition tables: {e}")
        return None
    table_df = table.lazy_frame.collect()
    table_df = table_df.with_columns(
        pl.lit(plate_name_from_url(url)).alias("plate_name"),
        pl.col("column").cast(pl.Int64),
        pl.col("path_in_well").cast(pl.Utf8),
    )

    required_columns = ["row", "column", "path_in_well"]
    for column in required_columns:
        if column not in table_df.columns:
            st.error(
                f"Condition table {table_name} does not contain required column {column}."
            )
            return None
    return table_df


@st.cache_data
def _collect_condition_table_from_plates_cached(
    list_urls: list[str],
    table_name: str,
) -> pl.DataFrame | None:
    """Load the condition table from the plate URLs."""
    condition_tables = []
    for url in list_urls:
        table_df = _load_single_plate_condition_table(url, table_name)
        if table_df is None:
            continue
        condition_tables.append(table_df)

    condition_table = pl.concat(condition_tables)
    return condition_table


@st.cache_data
def _collect_condition_table_from_images_cached(
    list_urls: list[str],
    table_name: str,
    mode: Literal["plate", "image"] = "plate",
) -> pl.DataFrame:
    """Load the condition table from the image URLs."""
    images = [get_ome_zarr_container(url, mode=mode) for url in list_urls]

    extras = [extras_from_url(url) for url in list_urls]
    # For more efficient loading, we should reimplement this
    # using the streamlit caches
    condition_table = asyncio.run(
        concatenate_image_tables_async(
            images=images,
            extras=extras,
            table_name=table_name,
        )
    )
    condition_table = condition_table.lazy_frame.collect()
    if mode == "plate":
        condition_table = condition_table.with_columns(
            pl.col("column").cast(pl.Int64),
            pl.col("path_in_well").cast(pl.Utf8),
        )
    return condition_table


def _join_setup_to_condition_table(
    plate_setup_df: pl.DataFrame,
    condition_df: pl.DataFrame,
    on=("plate_name", "row", "column", "path_in_well"),
) -> pl.DataFrame:
    """Join the condition table with the plate setup DataFrame."""
    plate_setup_df = plate_setup_df.join(
        condition_df,
        left_on=on,
        how="inner",
    )
    return plate_setup_df


def collect_condition_table_from_plates(
    plate_setup_df: pl.DataFrame,
    table_name: str,
) -> pl.DataFrame | None:
    """Load the condition table from the plate URLs."""
    plate_urls = plate_setup_df["plate_url"].unique().sort().to_list()
    condition_df = _collect_condition_table_from_plates_cached(plate_urls, table_name)
    if condition_df is None:
        return None
    return _join_setup_to_condition_table(plate_setup_df, condition_df)


def collect_condition_table_from_images(
    plate_setup_df: pl.DataFrame,
    table_name: str,
    on=("plate_name", "row", "column", "path_in_well"),
    mode: Literal["plate", "image"] = "plate",
) -> pl.DataFrame:
    """Load the condition table from the image URLs."""
    images_urls = plate_setup_df["image_url"].unique().sort().to_list()
    condition_df = _collect_condition_table_from_images_cached(
        images_urls, table_name, mode=mode
    )
    return _join_setup_to_condition_table(plate_setup_df, condition_df, on=on)


# ====================================================================
#
# Feature tables utils
#
# ====================================================================


@st.cache_data
def _load_plate_feature_table(
    url: str,
    table_name: str,
) -> pl.DataFrame:
    """Load the feature table from a single plate URL."""
    plate = get_ome_zarr_plate(url)
    try:
        table = plate.get_table_as(table_name, FeatureTable)
        reference_label = table.reference_label
    except Exception as e:
        st.error(f"Error loading feature tables: {e}")
        raise e

    table_df = table.lazy_frame.collect()
    table_df = table_df.with_columns(
        pl.lit(plate_name_from_url(url)).alias("plate_name"),
        pl.col("column").cast(pl.Int64),
        pl.col("path_in_well").cast(pl.Utf8),
        pl.lit(reference_label).alias("reference_label"),
    )

    required_columns = ["row", "column", "path_in_well"]
    for column in required_columns:
        if column not in table_df.columns:
            st.error(
                f"Feature table {table_name} does not contain required column {column}."
            )
            raise ValueError(
                f"Feature table {table_name} does not contain required column {column}."
            )

    return table_df


@st.cache_data
def _collect_feature_table_from_plates_cached(
    list_urls: list[str],
    table_name: str,
) -> pl.DataFrame:
    """Load the feature table from the plate URLs."""
    feature_tables = []
    for url in list_urls:
        table_df = _load_plate_feature_table(url, table_name)
        feature_tables.append(table_df)

    feature_table = pl.concat(feature_tables)
    return feature_table


@st.cache_data
def _collect_feature_table_from_images_cached(
    list_urls: list[str],
    table_name: str,
    mode: Literal["plate", "image"] = "plate",
) -> pl.DataFrame:
    """Load the feature table from the image URLs."""
    images = [get_ome_zarr_container(url, mode="plate") for url in list_urls]

    extras = [extras_from_url(url) for url in list_urls]
    # For more efficient loading, we should reimplement this
    # using the streamlit caches
    feature_table = asyncio.run(
        concatenate_image_tables_as_async(
            images=images,
            extras=extras,
            table_name=table_name,
            table_cls=FeatureTable,
            mode="lazy",
        )
    )
    feature_df = feature_table.lazy_frame.collect()
    if mode == "plate":
        feature_df = feature_df.with_columns(
            pl.col("column").cast(pl.Int64),
            pl.col("path_in_well").cast(pl.Utf8),
            pl.lit(feature_table.reference_label).alias("reference_label"),
        )
    else:
        feature_df = feature_df.with_columns(
            pl.lit(feature_table.reference_label).alias("reference_label"),
        )
    return feature_df


def _join_feature_table_to_setup(
    plate_setup_df: pl.DataFrame,
    feature_df: pl.DataFrame,
    on=("plate_name", "row", "column", "path_in_well"),
    drop=("plate_url",),
) -> pl.DataFrame:
    """Join the feature table with the plate setup DataFrame."""
    feature_df = feature_df.join(
        plate_setup_df,
        on=on,
        how="inner",
    )
    feature_df = feature_df.drop(drop)
    return feature_df


def collect_feature_table_from_plates(
    plate_setup_df: pl.DataFrame,
    table_name: str,
) -> pl.DataFrame | None:
    """Load the feature table from the plate URLs."""
    plate_urls = plate_setup_df["plate_url"].unique().sort().to_list()
    feature_table = _collect_feature_table_from_plates_cached(plate_urls, table_name)
    if feature_table is None:
        return None
    feature_table = _join_feature_table_to_setup(plate_setup_df, feature_table)
    return feature_table


def collect_feature_table_from_images(
    plate_setup_df: pl.DataFrame,
    table_name: str,
) -> pl.DataFrame:
    """Load the feature table from the image URLs."""
    images_urls = plate_setup_df["image_url"].unique().sort().to_list()
    feature_table = _collect_feature_table_from_images_cached(images_urls, table_name)
    feature_table = _join_feature_table_to_setup(plate_setup_df, feature_table)
    return feature_table
